/*
 * Copyright (c) 2009-2011 LodgON  (http://www.lodgon.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package com.lodgon.dali.core.ejb;

import com.lodgon.dali.core.Constants;
import com.lodgon.dali.core.entity.Content;
import com.lodgon.dali.core.entity.ContentPermission;
import com.lodgon.dali.core.entity.Field;
import com.lodgon.dali.core.entity.Group;
import com.lodgon.dali.core.entity.User;
import com.lodgon.dali.core.entity.search.ContentSearchParams;
import com.lodgon.dali.core.entity.search.ContentSearchParamsBuilder;
import com.lodgon.dali.core.entity.search.SearchParamsBuilder;
import com.lodgon.dali.core.search.ActionType;
import com.lodgon.dali.core.search.DefaultContentSearchProvider;
import com.lodgon.dali.core.search.EntityAction;
import com.lodgon.dali.core.search.SearchProvider;
import java.util.List;
import java.util.UUID;
import java.util.logging.Level;
import javax.annotation.PostConstruct;
import javax.ejb.EJB;
import javax.ejb.Stateless;
import javax.enterprise.event.Event;
import javax.enterprise.inject.Any;
import javax.enterprise.inject.Instance;
import javax.inject.Inject;
import javax.persistence.EntityManager;
import javax.persistence.NoResultException;
import javax.persistence.NonUniqueResultException;
import javax.persistence.PersistenceContext;
import javax.persistence.Query;

/**
 * This class provides the persistence operations for content.
 *
 * @author erwin
 */
@Stateless
public class ContentBean {
	@Inject @EntityAction(ActionType.CREATE)
	private Event<Content> createContentEvent;
	@Inject @EntityAction(ActionType.UPDATE)
	private Event<Content> updateContentEvent;
	@Inject @EntityAction(ActionType.DELETE)
	private Event<Content> deleteContentEvent;

  @PersistenceContext
  EntityManager em;

	@Inject @Any
	private Instance<SearchProvider<Content>> searchProviders;
	private SearchProvider<Content> searchProvider;

  @EJB
  AuthorizationBean authorizationBean;

	@PostConstruct
	private void initializeContentBean() {
		if (searchProviders != null) {
			if (! searchProviders.isUnsatisfied() && ! searchProviders.isAmbiguous()) {
				searchProvider = searchProviders.get();
			} else if (searchProviders.isAmbiguous()) {
				for (SearchProvider<Content> sp : searchProviders) {
					if (! (sp instanceof DefaultContentSearchProvider)) {
						searchProvider = sp;
					}
				}
			}
		}

		if (searchProvider == null) {
			Constants.LOGGER.log(Level.SEVERE, "No implementation for SearchProvider<Content> was found.");
		} else {
			Constants.LOGGER.log(Level.INFO, "Using {0} as SearchProvider<Content>.", searchProvider.getClass().getName());
		}
	}

	/**
   * Add a Content to persistent storage.
   * If <code>content.getUid()</code> is not <code>null</code>, the value of it should be unique.
   * If <code>content.getUid()</code> is null, this method will automatically generate a UUID.
	 *
	 * @param content the content that has to be made persistent
	 * @return the persisted Content object
	 * @throws DaliCoreException when a conflict occurs with the appKey or when the provided <code>content.getUid()</code> is not unique
	 */
  public Content create(Content content) throws DaliCoreException {
    if (content.getAppKey() == null) {
      content.setAppKey("");
    }
    if (content.getUid() != null) {
      Content existing = getByUid(content.getAppKey(), content.getUid());
      if (existing != null) {
        throw new DaliCoreException(DaliCoreException.UUID_CONFLICT, "content creation failed because "+content.getUid()+" already exists");
      }
    } else {
      content.setUid(UUID.randomUUID().toString());
    }
    if (content.getAuthor() != null && !content.getAppKey().equals(content.getAuthor().getAppKey())) {
      throw new DaliCoreException(DaliCoreException.APPKEY_CONFLICT, "appkey conflict in author "+content.getAuthor().getAppKey());
    }
    if (content.getGroup() != null && !content.getAppKey().equals(content.getGroup().getAppKey())) {
      throw new DaliCoreException(DaliCoreException.APPKEY_CONFLICT, "appkey conflict in group "+content.getGroup().getAppKey());
    }
    if (content.getParent() != null && !content.getAppKey().equals(content.getParent().getAppKey())) {
      throw new DaliCoreException(DaliCoreException.APPKEY_CONFLICT, "appkey conflict in parent "+content.getParent().getAppKey());
    }

    if (content.getFields() != null) {
      for (Field field : content.getFields()) {
        field.setContent(content);
      }
    }

    em.persist(content);
		createContentEvent.fire(content);
    return content;
  }

  /**
   * Updates a content in the persistent storage.
	 *
   * @param content the content to update
   * @return the updated Content object
	 * @throws DaliCoreException when a conflict occurs with the appKey or when the provided <code>content.getUid()</code> is not unique
   */
  public Content update(Content content) throws DaliCoreException {
    if (content.getAppKey() == null) {
      content.setAppKey("");
    }
    if (content.getUid() != null) {
      Content existing = getByUid(content.getAppKey(), content.getUid());
      if (existing != null) {
        if ((existing.getId() != null && ! existing.getId().equals(content.getId())) || content.getId() == null) {
          throw new DaliCoreException(DaliCoreException.UUID_CONFLICT, "content creation failed because "+content.getUid()+" already exists");
        }
      }
    } else {
      content.setUid(UUID.randomUUID().toString());
    }
    if (content.getAuthor() != null && !content.getAppKey().equals(content.getAuthor().getAppKey())) {
      throw new DaliCoreException(DaliCoreException.APPKEY_CONFLICT, "appkey conflict in author "+content.getAuthor().getAppKey());
    }
    if (content.getGroup() != null && !content.getAppKey().equals(content.getGroup().getAppKey())) {
      throw new DaliCoreException(DaliCoreException.APPKEY_CONFLICT, "appkey conflict in group "+content.getGroup().getAppKey());
    }
    if (content.getParent() != null && !content.getAppKey().equals(content.getParent().getAppKey())) {
      throw new DaliCoreException(DaliCoreException.APPKEY_CONFLICT, "appkey conflict in parent "+content.getParent().getAppKey());
    }

    if (content.getFields() != null) {
      for (Field f : content.getFields()) {
        f.setContent(content);
      }
    }

		Content dbcontent = em.find(Content.class, content.getId());
		String pathBefore = dbcontent.getPath();
		if (content.getParent() == null) {
			content.setPath("");
			content.setDepth(0);
		} else {
			content.setPath(content.getParent().getPath().isEmpty() ? "/" + content.getParent().getId() + "/" : content.getParent().getPath() + content.getParent().getId() + "/");
			content.setDepth(content.getParent().getDepth() + 1);
		}

		// update child content paths when path changed
		if (! pathBefore.equals(content.getPath())) {
			updateChildrenPaths(content);
		}

		content = em.merge(content);
		updateContentEvent.fire(content);
		return content;
  }

	private void updateChildrenPaths(Content parent) {
		List<Content> children = getByParent(parent.getAppKey(), parent.getId(), null, 1);
		for (Content child : children) {
			child.setPath(parent.getPath().isEmpty() ? "/" + parent.getId() + "/" : parent.getPath() + parent.getId() + "/");
			child.setDepth(parent.getDepth() + 1);
			updateChildrenPaths(child);
		}
	}

	/**
   * Remove a content from persistent storage. This is the same as calling
	 * <code>delete(contentId, false)</code>.
	 *
   * @param contentId the unique identifier of the content to remove
	 * @see #delete(int, boolean)
	 */
  public void delete(int contentId) {
		delete(em.find(Content.class, contentId));
  }

	/**
   * Remove a content from persistent storage. When cascade is set to true the
	 * entire tree below the content that is to be removed will be removed from
	 * persistent storage as well.
	 *
   * @param contentId the unique identifier of the content to remove
	 * @param cascade if set to true delete all child contents in a cascading way
	 * @see #delete(com.lodgon.dali.core.entity.Content, boolean)
	 */
	public void delete(int contentId, boolean cascade) {
		delete(em.find(Content.class, contentId), cascade);
  }

	/**
   * Remove a content from persistent storage. This is the same as calling
	 * <code>delete(content, false)</code>.
	 *
   * @param content the managed Content entity to remove
	 * @see #delete(com.lodgon.dali.core.entity.Content, boolean)
	 */
  public void delete(Content content) {
		delete(content, false);
	}

	/**
   * Remove a content from persistent storage. When cascade is set to true the
	 * entire tree below the content that is to be removed will be removed from
	 * persistent storage as well. In addition: all permissions that are linked
	 * with the content will be removed as well.
	 *
   * @param content the managed Content entity to remove
	 * @param cascade if set to true delete all child contents in a cascading way
	 */
	public void delete(Content content, boolean cascade) {
    if (content != null) {
	    List<ContentPermission> contentPermissions = authorizationBean.getContentPermissionsByContent(content);
		  for (ContentPermission contentPermission : contentPermissions) {
				em.remove(contentPermission);
			}

      List<Content> children = getByParent(content.getAppKey(), content, null, 1);
      for (Content child : children) {
				if (cascade) {
					delete(child, cascade);
				} else {
	        child.setParent(null);
					child.setPath("");
					child.setDepth(0);
		      updateChildrenPaths(em.merge(child));
				}
      }

			deleteContentEvent.fire(content);
      em.remove(content);
    }
  }

	/**
	 * Sets the position of the content to the newly specified position.
	 *
   * @param contentId the unique identifier of the content to update
	 * @param position the new position for the specified content
	 * @see #updatePosition(com.lodgon.dali.core.entity.Content, int)
   * @return the updated Content entity
	 */
	public Content updatePosition(int contentId, int position) {
		return updatePosition(em.find(Content.class, contentId), position);
	}

	/**
	 * Sets the position of the content to the newly specified position. This
	 * method serves as a convenience method to manage content that are ordered
	 * by their weight. Setting a new position for a certain content by using this
	 * method instead of {@link #update(com.lodgon.dali.core.entity.Content)} will
	 * automatically update all the weights of the sibling content entities. For
	 * example, assume we have five content entities with weight 1, 2, 3, 4 and 5
	 * respectively. When you change the position of the second content to 5,
	 * content number 3, 4 and 5 will automatically have their weight lowered by
	 * one.
	 *
	 * When creating a new content, you should set its weight value to the number
	 * of existing sibling content entities plus one. When deleting a content
	 * entity, you should first call this method with a position value of 0. This
	 * will remove the content entity from the ordered list.
	 *
	 * As a final note should be mentioned that the ordering that is implied here
	 * starts from position 1.
	 *
   * @param content the managed Content entity to update
	 * @param position the new position for the specified content
   * @return the updated Content entity
	 */
	public Content updatePosition(Content content, int position) {
		int currentPosition = content.getWeight();
		if (currentPosition == position) return content;

		if (content.getParent() == null) {
			if (position == 0) {
				Query q = em.createNamedQuery("Content.clearPosition");
				q.setParameter("appKey", content.getAppKey());
				q.setParameter("oldPosition", currentPosition);
				q.executeUpdate();
			} else if (currentPosition < position) {
				Query q = em.createNamedQuery("Content.higherPosition");
				q.setParameter("appKey", content.getAppKey());
				q.setParameter("oldPosition", currentPosition);
				q.setParameter("newPosition", position);
				q.executeUpdate();
			} else if (currentPosition > position) {
				Query q = em.createNamedQuery("Content.lowerPosition");
				q.setParameter("appKey", content.getAppKey());
				q.setParameter("oldPosition", currentPosition);
				q.setParameter("newPosition", position);
				q.executeUpdate();
			}
		} else {
			if (position == 0) {
				Query q = em.createNamedQuery("Content.clearPositionByParent");
				q.setParameter("appKey", content.getAppKey());
				q.setParameter("parent", content.getParent());
				q.setParameter("oldPosition", currentPosition);
				q.executeUpdate();
			} else if (currentPosition < position) {
				Query q = em.createNamedQuery("Content.higherPositionByParent");
				q.setParameter("appKey", content.getAppKey());
				q.setParameter("parent", content.getParent());
				q.setParameter("oldPosition", currentPosition);
				q.setParameter("newPosition", position);
				q.executeUpdate();
			} else if (currentPosition > position) {
				Query q = em.createNamedQuery("Content.lowerPositionByParent");
				q.setParameter("appKey", content.getAppKey());
				q.setParameter("parent", content.getParent());
				q.setParameter("oldPosition", currentPosition);
				q.setParameter("newPosition", position);
				q.executeUpdate();
			}
		}

		content.setWeight(position);
		return content;
	}

  /**
   * Find a content specified by the auto-created identifier.
	 *
   * @param contentId the identifier of the content
   * @return the Content that matches the specified identifier or null otherwise
   */
  public Content getById(int contentId) {
    return em.find(Content.class, contentId);
  }

  /**
   * Find a content specified by the unique identifier.
	 *
   * @param uid the unique identifier of the content
   * @return the Content that matches the specified uid or null otherwise
   */
  public Content getByUid(String uid) throws DaliCoreException {
    return getByUid("", uid);
  }

  /**
   * Find a content specified by the unique identifier within an application.
	 *
   * @param appKey the key of the application to search in
   * @param uid the unique identifier of the content
   * @return the Content that matches the specified uid in the application or null otherwise
   */
  public Content getByUid(String appKey, String uid) throws DaliCoreException {
    Query q = em.createNamedQuery("Content.findByUid");
    q.setParameter("uid", uid);
    q.setParameter("appKey", appKey);
    try {
      Content content = (Content) q.getSingleResult();
      return content;
    } catch (NoResultException e) {
			Constants.LOGGER.log(Level.INFO, "No content found with uid: {0}", uid);
    } catch (NonUniqueResultException e) {
			Constants.LOGGER.log(Level.SEVERE, "More than one content found with uid: {0}", uid);
			throw new DaliCoreException("non unique content result for " + uid);
    }
    return null;
  }

  /**
   * Count the number of contents that match a specific type.
	 *
   * @param type the type of the content
   * @return the number of contents that match the specified type
   */
  public Long countByType(int type) {
		return countSearch(SearchParamsBuilder.content().type(type).create());
  }

  /**
   * Count the number of contents that match a specific type within an
	 * application.
	 *
   * @param appKey the key of the application to search in
   * @param type the type of the content
   * @return the number of contents that match the specified type in the application
   */
  public Long countByType(String appKey, int type) {
		return countSearch(appKey, SearchParamsBuilder.content().type(type).create());
  }

	/**
	 * Find all contents that are of the specified type.
	 *
	 * @param type the type to search for
	 * @return All contents of the specified type
	 * @see #getByType(java.lang.String, int, java.lang.String, java.lang.String, java.lang.Integer, java.lang.Integer) 
	 */
	public List<Content> getByType(int type) {
		return search(SearchParamsBuilder.content().type(type).create());
	}

	/**
	 * Find all contents that are of the specified type within an application.
	 *
   * @param appKey the key of the application to search in
	 * @param type the type to search for
	 * @return All contents of the specified type in the application
	 * @see #getByType(java.lang.String, int, java.lang.String, java.lang.String, java.lang.Integer, java.lang.Integer) 
	 */
	public List<Content> getByType(String appKey, int type) {
		return search(appKey, SearchParamsBuilder.content().type(type).create());
	}

	/**
	 * Find a subset of contents that are of the specified type.
	 *
	 * @param type the type to search for
	 * @param sort on which field to sort the list of contents
	 * @param order in which order to perform the sort
	 * @param offset the start position of the first result, numbered from 0
	 * @param max the maximum number of results to retrieve
	 * @return A subset of contents of the specified type
	 * @see #getByType(java.lang.String, int, java.lang.String, java.lang.String, java.lang.Integer, java.lang.Integer) 
	 */
  public List<Content> getByType(int type, String sort, String order, Integer offset, Integer max) {
		ContentSearchParamsBuilder builder = SearchParamsBuilder.content().type(type).sort(sort).offset(offset).max(max);
		if ("desc".equalsIgnoreCase(order)) builder.desc();
		return search(builder.create());
  }

	/**
	 * Find a subset of contents that are of the specified type within an
	 * application. When offset is <code>null</code>, an offset of 0 will be used.
	 * When max is <code>null</code>, all results will be returned.<br/>
	 * Possible values for sort are: creationDate, modificationDate, title and weight.<br/>
	 * Possible values for order are: asc and desc.
	 *
   * @param appKey the key of the application to search in
	 * @param type the type to search for
	 * @param sort on which field to sort the list of contents
	 * @param order in which order to perform the sort
	 * @param offset the start position of the first result, numbered from 0
	 * @param max the maximum number of results to retrieve
	 * @return A subset of contents of the specified type in the application
	 */
  public List<Content> getByType(String appKey, int type, String sort, String order, Integer offset, Integer max) {
		ContentSearchParamsBuilder builder = SearchParamsBuilder.content().type(type).sort(sort).offset(offset).max(max);
		if ("desc".equalsIgnoreCase(order)) builder.desc();
		return search(appKey, builder.create());
  }

  /**
   * Count the number of contents of a parent content that match a specific type.
	 *
	 * @param parentId the identifier of the parent content
   * @param type the type of the content or <code>null</code> to search for all types
	 * @param depth how deep to search in the parent subtree
   * @return the number of contents of the parent content that match the specified type
	 * @see #countByParent(java.lang.String, com.lodgon.dali.core.entity.Content, java.lang.Integer, java.lang.Integer)
   */
  public Long countByParent(int parentId, Integer type, Integer depth) {
		ContentSearchParamsBuilder builder = SearchParamsBuilder.content().parentId(parentId).depth(depth);
		if (type != null) builder.type(type);
		return countSearch(builder.create());
  }

  /**
   * Count the number of contents of a parent content that match a specific
	 * type within an application.
	 *
   * @param appKey the key of the application to search in
	 * @param parentId the identifier of the parent content
   * @param type the type of the content or <code>null</code> to search for all types
	 * @param depth how deep to search in the parent subtree
   * @return the number of contents of the parent content that match the specified type in the application
	 * @see #countByParent(java.lang.String, com.lodgon.dali.core.entity.Content, java.lang.Integer, java.lang.Integer)
   */
  public Long countByParent(String appKey, int parentId, Integer type, Integer depth) {
		ContentSearchParamsBuilder builder = SearchParamsBuilder.content().parentId(parentId).depth(depth);
		if (type != null) builder.type(type);
		return countSearch(appKey, builder.create());
  }

  /**
   * Count the number of contents of a managed parent Content entity that match
	 * a specific type.
	 *
	 * @param parent the managed parent Content entity
   * @param type the type of the content or <code>null</code> to search for all types
	 * @param depth how deep to search in the parent subtree
   * @return the number of contents of the parent content that match the specified type
	 * @see #countByParent(java.lang.String, com.lodgon.dali.core.entity.Content, java.lang.Integer, java.lang.Integer)
   */
	public Long countByParent(Content parent, Integer type, Integer depth) {
		ContentSearchParamsBuilder builder = SearchParamsBuilder.content().parent(parent).depth(depth);
		if (type != null) builder.type(type);
		return countSearch(builder.create());
	}

  /**
   * Count the number of contents of a managed parent Content entity that match
	 * a specific type within an application. The <code>depth</code> parameter
	 * specifies how deep to search in the parent's subtree. A value 1 or <code>null</code>
	 * means that it will only return the count of all the children that are
	 * directly below the parent. A value of 2 will include their children and so
	 * on. A negative value will search the entire subtree.
	 *
   * @param appKey the key of the application to search in
	 * @param parent the managed parent Content entity
   * @param type the type of the content or <code>null</code> to search for all types
	 * @param depth how deep to search in the parent subtree
   * @return the number of contents of the parent content that match the specified type in the application
   */
	public Long countByParent(String appKey, Content parent, Integer type, Integer depth) {
		ContentSearchParamsBuilder builder = SearchParamsBuilder.content().parent(parent).depth(depth);
		if (type != null) builder.type(type);
		return countSearch(appKey, builder.create());
	}

	/**
	 * Find all child contents of a parent content that are of the specified type.
	 *
	 * @param parentId the identifier of the parent content
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @param depth how deep to search in the parent subtree
	 * @return All the children of the parent content that are of the specified type
	 * @see #getByParent(java.lang.String, com.lodgon.dali.core.entity.Content, java.lang.Integer, java.lang.Integer, java.lang.String, java.lang.String, java.lang.Integer, java.lang.Integer)
	 */
  public List<Content> getByParent(int parentId, Integer type, Integer depth) {
		ContentSearchParamsBuilder builder = SearchParamsBuilder.content().parentId(parentId).depth(depth);
		if (type != null) builder.type(type);
		return search(builder.create());
  }

	/**
	 * Find all child contents of a parent content that are of the specified type
	 * within an application.
	 *
   * @param appKey the key of the application to search in
	 * @param parentId the identifier of the parent content
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @param depth how deep to search in the parent subtree
	 * @return All the children of the parent content that are of the specified type in the application
	 * @see #getByParent(java.lang.String, com.lodgon.dali.core.entity.Content, java.lang.Integer, java.lang.Integer, java.lang.String, java.lang.String, java.lang.Integer, java.lang.Integer)
	 */
  public List<Content> getByParent(String appKey, int parentId, Integer type, Integer depth) {
		ContentSearchParamsBuilder builder = SearchParamsBuilder.content().parentId(parentId).depth(depth);
		if (type != null) builder.type(type);
		return search(appKey, builder.create());
  }

	/**
	 * Find all child contents of a managed parent Content entity that are of the
	 * specified type.
	 *
	 * @param parent the managed parent Content entity
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @param depth how deep to search in the parent subtree
	 * @return All the children of the parent content that are of the specified type
	 * @see #getByParent(java.lang.String, com.lodgon.dali.core.entity.Content, java.lang.Integer, java.lang.Integer, java.lang.String, java.lang.String, java.lang.Integer, java.lang.Integer)
	 */
  public List<Content> getByParent(Content parent, Integer type, Integer depth) {
		ContentSearchParamsBuilder builder = SearchParamsBuilder.content().parent(parent).depth(depth);
		if (type != null) builder.type(type);
		return search(builder.create());
  }

	/**
	 * Find all child contents of a managed parent Content entity that are of the
	 * specified type within an application.
	 *
   * @param appKey the key of the application to search in
	 * @param parent the managed parent Content entity
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @param depth how deep to search in the parent subtree
	 * @return All the children of the parent content that are of the specified type in the application
	 * @see #getByParent(java.lang.String, com.lodgon.dali.core.entity.Content, java.lang.Integer, java.lang.Integer, java.lang.String, java.lang.String, java.lang.Integer, java.lang.Integer)
	 */
  public List<Content> getByParent(String appKey, Content parent, Integer type, Integer depth) {
		ContentSearchParamsBuilder builder = SearchParamsBuilder.content().parent(parent).depth(depth);
		if (type != null) builder.type(type);
		return search(appKey, builder.create());
  }

	/**
	 * Find a subset of child contents of a parent content that are of the
	 * specified type.
	 *
	 * @param parentId the identifier of the parent content
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @param depth how deep to search in the parent subtree
	 * @param sort on which field to sort the list of contents
	 * @param order in which order to perform the sort
	 * @param offset the start position of the first result, numbered from 0
	 * @param max the maximum number of results to retrieve
	 * @return a subset of the children of the parent content that are of the specified type
	 * @see #getByParent(java.lang.String, com.lodgon.dali.core.entity.Content, java.lang.Integer, java.lang.Integer, java.lang.String, java.lang.String, java.lang.Integer, java.lang.Integer)
	 */
  public List<Content> getByParent(int parentId, Integer type, Integer depth, String sort, String order, Integer offset, Integer max) {
		ContentSearchParamsBuilder builder = SearchParamsBuilder.content().parentId(parentId).depth(depth).sort(sort).offset(offset).max(max);
		if (type != null) builder.type(type);
		if ("desc".equalsIgnoreCase(order)) builder.desc();
		return search(builder.create());
  }

	/**
	 * Find a subset of child contents of a parent content that are of the
	 * specified type within an application.
	 *
   * @param appKey the key of the application to search in
	 * @param parentId the identifier of the parent content
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @param depth how deep to search in the parent subtree
	 * @param sort on which field to sort the list of contents
	 * @param order in which order to perform the sort
	 * @param offset the start position of the first result, numbered from 0
	 * @param max the maximum number of results to retrieve
	 * @return a subset of the children of the parent content that are of the specified type in the application
	 * @see #getByParent(java.lang.String, com.lodgon.dali.core.entity.Content, java.lang.Integer, java.lang.Integer, java.lang.String, java.lang.String, java.lang.Integer, java.lang.Integer)
	 */
  public List<Content> getByParent(String appKey, int parentId, Integer type, Integer depth, String sort, String order, Integer offset, Integer max) {
		ContentSearchParamsBuilder builder = SearchParamsBuilder.content().parentId(parentId).depth(depth).sort(sort).offset(offset).max(max);
		if (type != null) builder.type(type);
		if ("desc".equalsIgnoreCase(order)) builder.desc();
		return search(appKey, builder.create());
  }

	/**
	 * Find a subset of child contents of a managed parent Content entity that
	 * are of the specified type.
	 *
	 * @param parent the managed parent Content entity
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @param depth how deep to search in the parent subtree
	 * @param sort on which field to sort the list of contents
	 * @param order in which order to perform the sort
	 * @param offset the start position of the first result, numbered from 0
	 * @param max the maximum number of results to retrieve
	 * @return a subset of the children of the parent content that are of the specified type
	 * @see #getByParent(java.lang.String, com.lodgon.dali.core.entity.Content, java.lang.Integer, java.lang.Integer, java.lang.String, java.lang.String, java.lang.Integer, java.lang.Integer)
	 */
  public List<Content> getByParent(Content parent, Integer type, Integer depth, String sort, String order, Integer offset, Integer max) {
		ContentSearchParamsBuilder builder = SearchParamsBuilder.content().parent(parent).depth(depth).sort(sort).offset(offset).max(max);
		if (type != null) builder.type(type);
		if ("desc".equalsIgnoreCase(order)) builder.desc();
		return search(builder.create());
  }

	/**
	 * Find a subset of child contents of a managed parent Content entity that
	 * are of the specified type within an application. When offset is
	 * <code>null</code>, an offset of 0 will be used. When max is
	 * <code>null</code>, all results will be returned.<br/>
	 * Possible values for sort are: creationDate, modificationDate, title and weight.<br/>
	 * Possible values for order are: asc and desc. The <code>depth</code> parameter
	 * specifies how deep to search in the parent's subtree. A value 1 or <code>null</code>
	 * means that it will only return the count of all the children that are
	 * directly below the parent. A value of 2 will include their children and so
	 * on. A negative value will search the entire subtree.
	 *
   * @param appKey the key of the application to search in
	 * @param parent the managed parent Content entity
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @param depth how deep to search in the parent subtree
	 * @param sort on which field to sort the list of contents
	 * @param order in which order to perform the sort
	 * @param offset the start position of the first result, numbered from 0
	 * @param max the maximum number of results to retrieve
	 * @return a subset of the children of the parent content that are of the specified type in the application
	 */
  public List<Content> getByParent(String appKey, Content parent, Integer type, Integer depth, String sort, String order, Integer offset, Integer max) {
		ContentSearchParamsBuilder builder = SearchParamsBuilder.content().parent(parent).depth(depth).sort(sort).offset(offset).max(max);
		if (type != null) builder.type(type);
		if ("desc".equalsIgnoreCase(order)) builder.desc();
		return search(appKey, builder.create());
  }

  /**
   * Count the number of contents of an author that match a specific type.
	 *
	 * @param authorId the identifier of the author
   * @param type the type of the content or <code>null</code> to search for all types
   * @return the number of contents of the author that match the specified type
   */
  public Long countByAuthor(int authorId, Integer type) {
		ContentSearchParamsBuilder builder = SearchParamsBuilder.content().authorId(authorId);
		if (type != null) builder.type(type);
		return countSearch(builder.create());
  }

  /**
   * Count the number of contents of an author that match a specific type
	 * within an application.
	 *
   * @param appKey the key of the application to search in
	 * @param authorId the identifier of the author
   * @param type the type of the content or <code>null</code> to search for all types
   * @return the number of contents of the author that match the specified type in the application
   */
  public Long countByAuthor(String appKey, int authorId, Integer type) {
		ContentSearchParamsBuilder builder = SearchParamsBuilder.content().authorId(authorId);
		if (type != null) builder.type(type);
		return countSearch(appKey, builder.create());
  }

  /**
   * Count the number of contents of a managed author entity that match a
	 * specific type.
	 *
	 * @param author the managed User entity
   * @param type the type of the content or <code>null</code> to search for all types
   * @return the number of contents of the author that match the specified type
   */
  public Long countByAuthor(User author, Integer type) {
		ContentSearchParamsBuilder builder = SearchParamsBuilder.content().author(author);
		if (type != null) builder.type(type);
		return countSearch(builder.create());
  }

  /**
   * Count the number of contents of a managed author entity that match a
	 * specific type within an application.
	 *
   * @param appKey the key of the application to search in
	 * @param author the managed User entity
   * @param type the type of the content or <code>null</code> to search for all types
   * @return the number of contents of the author that match the specified type in the application
   */
  public Long countByAuthor(String appKey, User author, Integer type) {
		ContentSearchParamsBuilder builder = SearchParamsBuilder.content().author(author);
		if (type != null) builder.type(type);
		return countSearch(appKey, builder.create());
  }

	/**
	 * Find all contents of an author that are of the specified type.
	 *
   * @param authorId the identifier of the author
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @return All contents of the author that are of the specified type
	 * @see #getByAuthor(java.lang.String, com.lodgon.dali.core.entity.User, java.lang.Integer, java.lang.String, java.lang.String, java.lang.Integer, java.lang.Integer)
	 */
  public List<Content> getByAuthor(int authorId, Integer type) {
		ContentSearchParamsBuilder builder = SearchParamsBuilder.content().authorId(authorId);
		if (type != null) builder.type(type);
		return search(builder.create());
  }

	/**
	 * Find all contents of an author that are of the specified type within an
	 * application.
	 *
   * @param appKey the key of the application to search in
   * @param authorId the identifier of the author
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @return All contents of the author that are of the specified type in the application
	 * @see #getByAuthor(java.lang.String, com.lodgon.dali.core.entity.User, java.lang.Integer, java.lang.String, java.lang.String, java.lang.Integer, java.lang.Integer)
	 */
  public List<Content> getByAuthor(String appKey, int authorId, Integer type) {
		ContentSearchParamsBuilder builder = SearchParamsBuilder.content().authorId(authorId);
		if (type != null) builder.type(type);
		return search(appKey, builder.create());
  }

	/**
	 * Find all contents of a managed author entity that are of the specified type.
	 *
   * @param author the managed User entity
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @return All contents of the author that are of the specified type
	 * @see #getByAuthor(java.lang.String, com.lodgon.dali.core.entity.User, java.lang.Integer, java.lang.String, java.lang.String, java.lang.Integer, java.lang.Integer)
	 */
  public List<Content> getByAuthor(User author, Integer type) {
		ContentSearchParamsBuilder builder = SearchParamsBuilder.content().author(author);
		if (type != null) builder.type(type);
		return search(builder.create());
  }

	/**
	 * Find all contents of a managed author entity that are of the specified type
	 * within an application.
	 *
   * @param appKey the key of the application to search in
   * @param author the managed User entity
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @return All contents of the author that are of the specified type in the application
	 * @see #getByAuthor(java.lang.String, com.lodgon.dali.core.entity.User, java.lang.Integer, java.lang.String, java.lang.String, java.lang.Integer, java.lang.Integer)
	 */
  public List<Content> getByAuthor(String appKey, User author, Integer type) {
		ContentSearchParamsBuilder builder = SearchParamsBuilder.content().author(author);
		if (type != null) builder.type(type);
		return search(appKey, builder.create());
  }

	/**
	 * Find a subset of contents of an author that are of the specified type.
	 *
   * @param authorId the identifier of the author
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @param sort on which field to sort the list of contents
	 * @param order in which order to perform the sort
	 * @param offset the start position of the first result, numbered from 0
	 * @param max the maximum number of results to retrieve
	 * @return A subset of contents of the author that are of the specified type
	 * @see #getByAuthor(java.lang.String, com.lodgon.dali.core.entity.User, java.lang.Integer, java.lang.String, java.lang.String, java.lang.Integer, java.lang.Integer)
	 */
  public List<Content> getByAuthor(int authorId, Integer type, String sort, String order, Integer offset, Integer max) {
		ContentSearchParamsBuilder builder = SearchParamsBuilder.content().authorId(authorId).sort(sort).offset(offset).max(max);
		if (type != null) builder.type(type);
		if ("desc".equalsIgnoreCase(order)) builder.desc();
		return search(builder.create());
  }

	/**
	 * Find a subset of contents of an author that are of the specified type
	 * within an application.
	 *
   * @param appKey the key of the application to search in
   * @param authorId the identifier of the author
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @param sort on which field to sort the list of contents
	 * @param order in which order to perform the sort
	 * @param offset the start position of the first result, numbered from 0
	 * @param max the maximum number of results to retrieve
	 * @return A subset of contents of the author that are of the specified type in the application
	 * @see #getByAuthor(java.lang.String, com.lodgon.dali.core.entity.User, java.lang.Integer, java.lang.String, java.lang.String, java.lang.Integer, java.lang.Integer)
	 */
  public List<Content> getByAuthor(String appKey, int authorId, Integer type, String sort, String order, Integer offset, Integer max) {
		ContentSearchParamsBuilder builder = SearchParamsBuilder.content().authorId(authorId).sort(sort).offset(offset).max(max);
		if (type != null) builder.type(type);
		if ("desc".equalsIgnoreCase(order)) builder.desc();
		return search(appKey, builder.create());
  }

	/**
	 * Find a subset of contents of a managed author entity that are of the
	 * specified type.
	 *
	 * @param author the managed User entity
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @param sort on which field to sort the list of contents
	 * @param order in which order to perform the sort
	 * @param offset the start position of the first result, numbered from 0
	 * @param max the maximum number of results to retrieve
	 * @return A subset of contents of the author that are of the specified type
	 * @see #getByAuthor(java.lang.String, com.lodgon.dali.core.entity.User, java.lang.Integer, java.lang.String, java.lang.String, java.lang.Integer, java.lang.Integer)
	 */
  public List<Content> getByAuthor(User author, Integer type, String sort, String order, Integer offset, Integer max) {
		ContentSearchParamsBuilder builder = SearchParamsBuilder.content().author(author).sort(sort).offset(offset).max(max);
		if (type != null) builder.type(type);
		if ("desc".equalsIgnoreCase(order)) builder.desc();
		return search(builder.create());
  }

	/**
	 * Find a subset of contents of a managed author entity that are of the
	 * specified type within an application. When offset is <code>null</code>, an
	 * offset of 0 will be used. When max is <code>null</code>, all results will
	 * be returned.<br/>
	 * Possible values for sort are: creationDate, modificationDate, title and weight.<br/>
	 * Possible values for order are: asc and desc.
	 *
   * @param appKey the key of the application to search in
	 * @param author the managed User entity
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @param sort on which field to sort the list of contents
	 * @param order in which order to perform the sort
	 * @param offset the start position of the first result, numbered from 0
	 * @param max the maximum number of results to retrieve
	 * @return A subset of contents of the author that are of the specified type in the application
	 */
  public List<Content> getByAuthor(String appKey, User author, Integer type, String sort, String order, Integer offset, Integer max) {
		ContentSearchParamsBuilder builder = SearchParamsBuilder.content().author(author).sort(sort).offset(offset).max(max);
		if (type != null) builder.type(type);
		if ("desc".equalsIgnoreCase(order)) builder.desc();
		return search(appKey, builder.create());
  }

  /**
   * Count the number of contents of a group that match a specific type.
	 *
	 * @param groupId the identifier of the group
   * @param type the type of the content or <code>null</code> to search for all types
   * @return the number of contents of the group that match the specified type
   */
  public Long countByGroup(int groupId, Integer type) {
		ContentSearchParamsBuilder builder = SearchParamsBuilder.content().groupId(groupId);
		if (type != null) builder.type(type);
		return countSearch(builder.create());
  }

  /**
   * Count the number of contents of a group that match a specific type within
	 * an application.
	 *
   * @param appKey the key of the application to search in
	 * @param groupId the identifier of the group
   * @param type the type of the content or <code>null</code> to search for all types
   * @return the number of contents of the group that match the specified type in the application
   */
  public Long countByGroup(String appKey, int groupId, Integer type) {
		ContentSearchParamsBuilder builder = SearchParamsBuilder.content().groupId(groupId);
		if (type != null) builder.type(type);
		return countSearch(appKey, builder.create());
  }

  /**
   * Count the number of contents of a managed Group entity that match a
	 * specific type.
	 *
	 * @param group the managed Group entity
   * @param type the type of the content or <code>null</code> to search for all types
   * @return the number of contents of the group that match the specified type
   */
  public Long countByGroup(Group group, Integer type) {
		ContentSearchParamsBuilder builder = SearchParamsBuilder.content().group(group);
		if (type != null) builder.type(type);
		return countSearch(builder.create());
  }

  /**
   * Count the number of contents of a managed Group entity that match a
	 * specific type within an application.
	 *
   * @param appKey the key of the application to search in
	 * @param group the managed Group entity
   * @param type the type of the content or <code>null</code> to search for all types
   * @return the number of contents of the group that match the specified type in the application
   */
  public Long countByGroup(String appKey, Group group, Integer type) {
		ContentSearchParamsBuilder builder = SearchParamsBuilder.content().group(group);
		if (type != null) builder.type(type);
		return countSearch(appKey, builder.create());
  }

	/**
	 * Find all contents of a group that are of the specified type.
	 *
   * @param groupId the identifier of the group
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @return All contents of the group that are of the specified type
	 * @see #getByGroup(java.lang.String, com.lodgon.dali.core.entity.Group, java.lang.Integer, java.lang.String, java.lang.String, java.lang.Integer, java.lang.Integer)
	 */
  public List<Content> getByGroup(int groupId, Integer type) {
		ContentSearchParamsBuilder builder = SearchParamsBuilder.content().groupId(groupId);
		if (type != null) builder.type(type);
		return search(builder.create());
  }

	/**
	 * Find all contents of a group that are of the specified type within an
	 * application.
	 *
   * @param appKey the key of the application to search in
   * @param groupId the identifier of the group
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @return All contents of the group that are of the specified type in the application
	 * @see #getByGroup(java.lang.String, com.lodgon.dali.core.entity.Group, java.lang.Integer, java.lang.String, java.lang.String, java.lang.Integer, java.lang.Integer)
	 */
  public List<Content> getByGroup(String appKey, int groupId, Integer type) {
		ContentSearchParamsBuilder builder = SearchParamsBuilder.content().groupId(groupId);
		if (type != null) builder.type(type);
		return search(appKey, builder.create());
  }

	/**
	 * Find all contents of a managed Group entity that are of the specified type.
	 *
   * @param group the managed Group entity
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @return All contents of the group that are of the specified type
	 * @see #getByGroup(java.lang.String, com.lodgon.dali.core.entity.Group, java.lang.Integer, java.lang.String, java.lang.String, java.lang.Integer, java.lang.Integer)
	 */
  public List<Content> getByGroup(Group group, Integer type) {
		ContentSearchParamsBuilder builder = SearchParamsBuilder.content().group(group);
		if (type != null) builder.type(type);
		return search(builder.create());
  }

	/**
	 * Find all contents of a managed Group entity that are of the specified type
	 * within an application.
	 *
   * @param appKey the key of the application to search in
   * @param group the managed Group entity
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @return All contents of the group that are of the specified type in the application
	 * @see #getByGroup(java.lang.String, com.lodgon.dali.core.entity.Group, java.lang.Integer, java.lang.String, java.lang.String, java.lang.Integer, java.lang.Integer)
	 */
  public List<Content> getByGroup(String appKey, Group group, Integer type) {
		ContentSearchParamsBuilder builder = SearchParamsBuilder.content().group(group);
		if (type != null) builder.type(type);
		return search(appKey, builder.create());
  }

	/**
	 * Find a subset of contents of a group that are of the specified type.
	 *
   * @param groupId the identifier of the group
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @param sort on which field to sort the list of contents
	 * @param order in which order to perform the sort
	 * @param offset the start position of the first result, numbered from 0
	 * @param max the maximum number of results to retrieve
	 * @return A subset of contents of the group that are of the specified type
	 * @see #getByGroup(java.lang.String, com.lodgon.dali.core.entity.Group, java.lang.Integer, java.lang.String, java.lang.String, java.lang.Integer, java.lang.Integer)
	 */
  public List<Content> getByGroup(int groupId, Integer type, String sort, String order, Integer offset, Integer max) {
		ContentSearchParamsBuilder builder = SearchParamsBuilder.content().groupId(groupId).sort(sort).offset(offset).max(max);
		if (type != null) builder.type(type);
		if ("desc".equalsIgnoreCase(order)) builder.desc();
		return search(builder.create());
  }

	/**
	 * Find a subset of contents of a group that are of the specified type
	 * within an application.
	 *
   * @param appKey the key of the application to search in
   * @param groupId the identifier of the group
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @param sort on which field to sort the list of contents
	 * @param order in which order to perform the sort
	 * @param offset the start position of the first result, numbered from 0
	 * @param max the maximum number of results to retrieve
	 * @return A subset of contents of the group that are of the specified type in the application
	 * @see #getByGroup(java.lang.String, com.lodgon.dali.core.entity.Group, java.lang.Integer, java.lang.String, java.lang.String, java.lang.Integer, java.lang.Integer)
	 */
  public List<Content> getByGroup(String appKey, int groupId, Integer type, String sort, String order, Integer offset, Integer max) {
		ContentSearchParamsBuilder builder = SearchParamsBuilder.content().groupId(groupId).sort(sort).offset(offset).max(max);
		if (type != null) builder.type(type);
		if ("desc".equalsIgnoreCase(order)) builder.desc();
		return search(appKey, builder.create());
  }

	/**
	 * Find a subset of contents of a managed Group entity that are of the
	 * specified type.
	 *
   * @param group the managed Group entity
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @param sort on which field to sort the list of contents
	 * @param order in which order to perform the sort
	 * @param offset the start position of the first result, numbered from 0
	 * @param max the maximum number of results to retrieve
	 * @return A subset of contents of the group that are of the specified type
	 * @see #getByGroup(java.lang.String, com.lodgon.dali.core.entity.Group, java.lang.Integer, java.lang.String, java.lang.String, java.lang.Integer, java.lang.Integer)
	 */
  public List<Content> getByGroup(Group group, Integer type, String sort, String order, Integer offset, Integer max) {
		ContentSearchParamsBuilder builder = SearchParamsBuilder.content().group(group).sort(sort).offset(offset).max(max);
		if (type != null) builder.type(type);
		if ("desc".equalsIgnoreCase(order)) builder.desc();
		return search(builder.create());
  }

	/**
	 * Find a subset of contents of a managed Group entity that are of the
	 * specified type within an application. When offset is <code>null</code>, an
	 * offset of 0 will be used. When max is <code>null</code>, all results will
	 * be returned.<br/>
	 * Possible values for sort are: creationDate, modificationDate, title and weight.<br/>
	 * Possible values for order are: asc and desc.
	 *
   * @param appKey the key of the application to search in
   * @param group the managed Group entity
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @param sort on which field to sort the list of contents
	 * @param order in which order to perform the sort
	 * @param offset the start position of the first result, numbered from 0
	 * @param max the maximum number of results to retrieve
	 * @return A subset of contents of the group that are of the specified type in the application
	 */
  public List<Content> getByGroup(String appKey, Group group, Integer type, String sort, String order, Integer offset, Integer max) {
		ContentSearchParamsBuilder builder = SearchParamsBuilder.content().group(group).sort(sort).offset(offset).max(max);
		if (type != null) builder.type(type);
		if ("desc".equalsIgnoreCase(order)) builder.desc();
		return search(appKey, builder.create());
  }

  /**
   * Count the total number of contents.
	 *
   * @return the total number of contents
   */
	public Long countAll() {
		return countAll("");
	}

  /**
   * Count the total number of contents within an application.
	 *
   * @param appKey the key of the application to search in
   * @return the total number of contents in the application
   */
	public Long countAll(String appKey) {
		Query q = em.createNamedQuery("Content.count");
		q.setParameter("appKey", appKey);
		return (Long) q.getSingleResult();
	}

  /**
   * Count the number of contents that match the specified search parameters.
	 *
	 * @param params the search parameters that are used to generate the search query
   * @return the total number of contents that match the search
   */
  public Long countSearch(ContentSearchParams params) {
    return countSearch("", params);
  }

  /**
   * Count the number of contents that match the specified search parameters
	 * within an application.
	 *
   * @param appKey the key of the application to search in
	 * @param params the search parameters that are used to generate the search query
   * @return the total number of contents that match the search in the application
   */
  public Long countSearch(String appKey, ContentSearchParams params) {
		return searchProvider.count(appKey, params);
  }

	/**
	 * Finds all contents that match the specified search parameters.
	 *
	 * @param params the search parameters that are used to generate the search query
	 * @return all contents that match the specified search parameters in an application
	 */
  public List<Content> search(ContentSearchParams params) {
		return search("", params);
  }

	/**
	 * Finds all contents that match the specified search parameters within an
	 * application.
	 *
   * @param appKey the key of the application to search in
	 * @param params the search parameters that are used to generate the search query
	 * @return all contents that match the specified search parameters in an application
	 */
  public List<Content> search(String appKey, ContentSearchParams params) {
		return searchProvider.search(appKey, params);
  }
}
