/*
 * Copyright (c) 2009-2011 LodgON  (http://www.lodgon.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package com.lodgon.dali.core.ejb;

import com.lodgon.dali.core.Constants;
import com.lodgon.dali.core.entity.Field;
import com.lodgon.dali.core.entity.Group;
import com.lodgon.dali.core.entity.GroupPermission;
import com.lodgon.dali.core.entity.TypePermission;
import com.lodgon.dali.core.entity.User;
import com.lodgon.dali.core.entity.search.GroupSearchParams;
import com.lodgon.dali.core.entity.search.GroupSearchParamsBuilder;
import com.lodgon.dali.core.entity.search.SearchParamsBuilder;
import com.lodgon.dali.core.search.ActionType;
import com.lodgon.dali.core.search.DefaultGroupSearchProvider;
import com.lodgon.dali.core.search.EntityAction;
import com.lodgon.dali.core.search.SearchProvider;
import java.util.List;
import java.util.UUID;
import java.util.logging.Level;
import javax.annotation.PostConstruct;
import javax.ejb.EJB;
import javax.ejb.Stateless;
import javax.enterprise.event.Event;
import javax.enterprise.inject.Any;
import javax.enterprise.inject.Instance;
import javax.inject.Inject;
import javax.persistence.EntityManager;
import javax.persistence.NoResultException;
import javax.persistence.NonUniqueResultException;
import javax.persistence.PersistenceContext;
import javax.persistence.Query;

/**
 * This class provides the persistence operations for groups.
 *
 * @author erwin
 */
@Stateless
public class GroupBean {
	@Inject @EntityAction(ActionType.CREATE)
	private Event<Group> createGroupEvent;
	@Inject @EntityAction(ActionType.UPDATE)
	private Event<Group> updateGroupEvent;
	@Inject @EntityAction(ActionType.DELETE)
	private Event<Group> deleteGroupEvent;

	@EJB
  AuthorizationBean authorizationBean;

	@Inject @Any
	private Instance<SearchProvider<Group>> searchProviders;
	private SearchProvider<Group> searchProvider;

	@PersistenceContext
  EntityManager em;

	@PostConstruct
	private void initializeContentBean() {
		if (searchProviders != null) {
			if (! searchProviders.isUnsatisfied() && ! searchProviders.isAmbiguous()) {
				searchProvider = searchProviders.get();
			} else if (searchProviders.isAmbiguous()) {
				for (SearchProvider<Group> sp : searchProviders) {
					if (! (sp instanceof DefaultGroupSearchProvider)) {
						searchProvider = sp;
					}
				}
			}
		}

		if (searchProvider == null) {
			Constants.LOGGER.log(Level.SEVERE, "No implementation for SearchProvider<Group> was found.");
		} else {
			Constants.LOGGER.log(Level.INFO, "Using {0} as SearchProvider<Group>.", searchProvider.getClass().getName());
		}
	}

	/**
   * Add a Group to persistent storage.
   * If <code>group.getUid()</code> is not <code>null</code>, the value of it should be unique.
   * If <code>group.getUid()</code> is null, this method will automatically generate a UUID.
	 *
	 * @param group the group that has to be made persistent
	 * @return the persisted Group object
	 * @throws DaliCoreException when a conflict occurs with the appKey or when the provided <code>group.getUid()</code> is not unique
	 */
  public Group create(Group group) throws DaliCoreException {
    if (group.getAppKey() == null) {
      group.setAppKey("");
    }
    if (group.getAuthor() != null && !group.getAppKey().equals(group.getAuthor().getAppKey())) {
      throw new DaliCoreException(DaliCoreException.APPKEY_CONFLICT, "appkey conflict in author "+group.getAuthor().getAppKey());
    }
    if (group.getParent() != null && !group.getAppKey().equals(group.getParent().getAppKey())) {
      throw new DaliCoreException(DaliCoreException.APPKEY_CONFLICT, "appkey conflict in parent "+group.getParent().getAppKey());
    }

    if (group.getFields() != null) {
      for (Field field : group.getFields()) {
        field.setGroup(group);
      }
    }
    if (group.getUid() != null) {
      Group existing = getByUid(group.getAppKey(), group.getUid());
      if (existing != null) {
        throw new DaliCoreException(DaliCoreException.UUID_CONFLICT, "group creation failed because " + group.getUid() + " already exists");
      }
    } else {
      group.setUid(UUID.randomUUID().toString());
    }

		em.persist(group);
		createGroupEvent.fire(group);
    return group;
  }

  /**
   * Updates a group in the persistent storage.
	 *
   * @param group the group to update
   * @return the updated Group object
   * @throws DaliCoreException when a conflict occurs with the appKey
   */
  public Group update(Group group) throws DaliCoreException {
    if (group.getAppKey() == null) {
      group.setAppKey("");
    }
    if (group.getAuthor() != null && !group.getAppKey().equals(group.getAuthor().getAppKey())) {
      throw new DaliCoreException(DaliCoreException.APPKEY_CONFLICT, "appkey conflict in author "+group.getAuthor().getAppKey());
    }
    if (group.getParent() != null && !group.getAppKey().equals(group.getParent().getAppKey())) {
      throw new DaliCoreException(DaliCoreException.APPKEY_CONFLICT, "appkey conflict in parent "+group.getParent().getAppKey());
    }

    if (group.getFields() != null) {
      for (Field f : group.getFields()) {
        f.setGroup(group);
      }
    }

		Group dbgroup = em.find(Group.class, group.getId());
    if (dbgroup.getMembers() != null) {
      group.setMembers(dbgroup.getMembers());
    }

		String pathBefore = dbgroup.getPath();
		if (group.getParent() == null) {
			group.setPath("");
			group.setDepth(0);
		} else {
			group.setPath(group.getParent().getPath().isEmpty() ? "/" + group.getParent().getId() + "/" : group.getParent().getPath() + group.getParent().getId() + "/");
			group.setDepth(group.getParent().getDepth() + 1);
		}

		// update child group paths when path changed
		if (! pathBefore.equals(group.getPath())) {
			updateChildrenPaths(group);
		}

    group = em.merge(group);
		updateGroupEvent.fire(group);
		return group;
  }

	private void updateChildrenPaths(Group parent) {
		List<Group> children = getByParent(parent.getAppKey(), parent.getId(), null, 1);
		for (Group child : children) {
			child.setPath(parent.getPath().isEmpty() ? "/" + parent.getId() + "/" : parent.getPath() + parent.getId() + "/");
			child.setDepth(parent.getDepth() + 1);
			updateChildrenPaths(child);
		}
	}

	/**
   * Remove a group from persistent storage. This is the same as calling
	 * <code>delete(groupId, false)</code>.
	 *
   * @param groupId the unique identifier of the group to remove
	 * @see #delete(int, boolean)
	 */
  public void delete(int groupId) {
    delete(em.find(Group.class, groupId));
  }

	/**
   * Remove a group from persistent storage. When cascade is set to true the
	 * entire tree below the group that is to be removed will be removed from
	 * persistent storage as well.
	 *
   * @param groupId the unique identifier of the group to remove
	 * @param cascade if set to true delete all child groups in a cascading way
	 * @see #delete(com.lodgon.dali.core.entity.Group, boolean)
	 */
  public void delete(int groupId, boolean cascade) {
    delete(em.find(Group.class, groupId), cascade);
  }

	/**
   * Remove a group from persistent storage. This is the same as calling
	 * <code>delete(group, false)</code>.
	 *
   * @param group the managed Group entity to remove
	 * @see #delete(com.lodgon.dali.core.entity.Group, boolean)
	 */
  public void delete(Group group) {
		delete(group, false);
	}

	/**
   * Remove a group from persistent storage. When cascade is set to true the
	 * entire tree below the group that is to be removed will be removed from
	 * persistent storage as well. In addition: all permissions that are linked
	 * with the group will be removed and all members will be removed from the
	 * group prior to its removal.
	 *
   * @param group the managed Group entity to remove
	 * @param cascade if set to true delete all child groups in a cascading way
	 */
  public void delete(Group group, boolean cascade) {
    if (group != null) {
	    List<TypePermission> typePermissions = authorizationBean.getTypePermissionsByGroup(group);
		  for (TypePermission typePermission : typePermissions) {
				em.remove(typePermission);
			}
	    List<GroupPermission> groupPermissions = authorizationBean.getGroupPermissionsByGroup(group);
		  for (GroupPermission groupPermission : groupPermissions) {
				em.remove(groupPermission);
			}
	    List<GroupPermission> targetGroupPermissions = authorizationBean.getGroupPermissionsByTarget(group);
		  for (GroupPermission targetGroupPermission : targetGroupPermissions) {
				em.remove(targetGroupPermission);
			}

			List<Group> children = getByParent(group.getAppKey(), group, null, 1);
			for (Group child : children) {
				if (cascade) {
					delete(child, cascade);
				} else {
					child.setParent(null);
					child.setPath("");
					child.setDepth(0);
					updateChildrenPaths(em.merge(child));
				}
			}

			List<User> members = group.getMembers();
			for (User member : members) {
				member.getGroups().remove(group);
			}
			group.getMembers().clear();

			deleteGroupEvent.fire(group);
			em.remove(group);
		}
  }

  /**
   * Find a group specified by the auto-created identifier.
	 *
   * @param groupId the identifier of the group
   * @return the Group that matches the specified identifier or null otherwise
   */
  public Group getById(int groupId) {
    return em.find(Group.class, groupId);
  }

  /**
   * Find a group specified by the unique identifier.
	 *
   * @param uid the unique identifier of the group
   * @return the Group that matches the specified uid or null otherwise
   */
	public Group getByUid(String uid) throws DaliCoreException {
		return getByUid("", uid);
	}

  /**
   * Find a group specified by the unique identifier within an application.
	 *
   * @param appKey the key of the application to search in
   * @param uid the unique identifier of the group
   * @return the Group that matches the specified uid in the application or null otherwise
   */
  public Group getByUid(String appKey, String uid) throws DaliCoreException {
    Query q = em.createNamedQuery("Group.findByUid");
    q.setParameter("uid", uid);
    q.setParameter("appKey", appKey);
    try {
      Group group = (Group) q.getSingleResult();
      return group;
    } catch (NoResultException e) {
			Constants.LOGGER.log(Level.INFO, "No group found with uid: {0}", uid);
    } catch (NonUniqueResultException nue) {
			Constants.LOGGER.log(Level.SEVERE, "More than one group found with uid: {0}", uid);
			throw new DaliCoreException("non unique group result for " + uid);
    }
    return null;
  }

  /**
   * Count the number of groups that match a specific type.
	 *
   * @param type the type of the group
   * @return the number of groups that match the specified type
   */
  public Long countByType(int type) {
		return countSearch(SearchParamsBuilder.group().type(type).create());
  }

  /**
   * Count the number of groups that match a specific type within an
	 * application.
	 *
   * @param appKey the key of the application to search in
   * @param type the type of the group
   * @return the number of groups that match the specified type in the application
   */
  public Long countByType(String appKey, int type) {
		return countSearch(appKey, SearchParamsBuilder.group().type(type).create());
  }

	/**
	 * Find all groups that are of the specified type.
	 *
	 * @param type the type to search for
	 * @return All groups of the specified type
	 * @see #getByType(java.lang.String, int, java.lang.Integer, java.lang.Integer)
	 */
  public List<Group> getByType(int type) {
		return search(SearchParamsBuilder.group().type(type).create());
  }

	/**
	 * Find all groups that are of the specified type within an application.
	 *
   * @param appKey the key of the application to search in
	 * @param type the type to search for
	 * @return All groups of the specified type in the application
	 * @see #getByType(java.lang.String, int, java.lang.Integer, java.lang.Integer)
	 */
  public List<Group> getByType(String appKey, int type) {
		return search(appKey, SearchParamsBuilder.group().type(type).create());
  }

	/**
	 * Find a subset of groups that are of the specified type.
	 *
	 * @param type the type to search for
	 * @param offset the start position of the first result, numbered from 0
	 * @param max the maximum number of results to retrieve
	 * @return A subset of groups of the specified type
	 * @see #getByType(java.lang.String, int, java.lang.Integer, java.lang.Integer)
	 */
  public List<Group> getByType(int type, Integer offset, Integer max) {
		return search(SearchParamsBuilder.group().type(type).offset(offset).max(max).create());
  }

	/**
	 * Find a subset of groups that are of the specified type within an
	 * application. When offset is <code>null</code>, an offset of 0 will be used.
	 * When max is <code>null</code>, all results will be returned.
	 *
   * @param appKey the key of the application to search in
	 * @param type the type to search for
	 * @param offset the start position of the first result, numbered from 0
	 * @param max the maximum number of results to retrieve
	 * @return A subset of groups of the specified type in the application
	 */
  public List<Group> getByType(String appKey, int type, Integer offset, Integer max) {
		return search(appKey, SearchParamsBuilder.group().type(type).offset(offset).max(max).create());
  }

	/**
	 * Count the number of groups of the specified type where a user belongs to.
	 *
	 * @param memberId the identifier of the user
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @return the number of groups of the specified type where the user belongs to
	 */
	public Long countByMember(int memberId, Integer type) {
		GroupSearchParamsBuilder builder = SearchParamsBuilder.group().memberId(memberId);
		if (type != null) builder.type(type);
		return countSearch(builder.create());
	}

	/**
	 * Count the number of groups of the specified type where a user belongs to
	 * within an application.
	 *
   * @param appKey the key of the application to search in
	 * @param memberId the identifier of the user
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @return the number of groups of the specified type where the user belongs to in the application
	 */
	public Long countByMember(String appKey, int memberId, Integer type) {
		GroupSearchParamsBuilder builder = SearchParamsBuilder.group().memberId(memberId);
		if (type != null) builder.type(type);
		return countSearch(appKey, builder.create());
	}

	/**
	 * Count the number of groups of the specified type where a managed User
	 * entity belongs to.
	 *
	 * @param member the managed User entity
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @return the number of groups of the specified type where the user belongs to
	 */
	public Long countByMember(User member, Integer type) {
		GroupSearchParamsBuilder builder = SearchParamsBuilder.group().member(member);
		if (type != null) builder.type(type);
		return countSearch(builder.create());
	}

	/**
	 * Count the number of groups of the specified type where a managed User
	 * entity belongs to within an application.
	 *
   * @param appKey the key of the application to search in
	 * @param member the managed User entity
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @return the number of groups of the specified type where the user belongs to in the application
	 */
	public Long countByMember(String appKey, User member, Integer type) {
		GroupSearchParamsBuilder builder = SearchParamsBuilder.group().member(member);
		if (type != null) builder.type(type);
		return countSearch(appKey, builder.create());
	}

	/**
	 * Find all groups of the specified type where a user belongs to.
	 *
	 * @param memberId the identifier of the user
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @return All groups of the specified type where the user belongs to
	 * @see #getByMember(java.lang.String, int, java.lang.Integer)
	 */
  public List<Group> getByMember(int memberId, Integer type) {
		GroupSearchParamsBuilder builder = SearchParamsBuilder.group().memberId(memberId);
		if (type != null) builder.type(type);
		return search(builder.create());
  }

	/**
	 * Find all groups of the specified type where a user belongs to within an
	 * application.
	 *
   * @param appKey the key of the application to search in
	 * @param memberId the identifier of the user
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @return All groups of the specified type where the user belongs to in the application
	 */
  public List<Group> getByMember(String appKey, int memberId, Integer type) {
		GroupSearchParamsBuilder builder = SearchParamsBuilder.group().memberId(memberId);
		if (type != null) builder.type(type);
		return search(appKey, builder.create());
  }

	/**
	 * Find all groups of the specified type where a managed User entity
	 * belongs to.
	 *
	 * @param member the managed User entity
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @return All groups of the specified type where the user belongs to
	 * @see #getByMember(java.lang.String, com.lodgon.dali.core.entity.User, java.lang.Integer)
	 */
  public List<Group> getByMember(User member, Integer type) {
		GroupSearchParamsBuilder builder = SearchParamsBuilder.group().member(member);
		if (type != null) builder.type(type);
		return search(builder.create());
  }

	/**
	 * Find all groups of the specified type where a managed User entity
	 * belongs to within an application.
	 *
   * @param appKey the key of the application to search in
	 * @param member the managed User entity
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @return All groups of the specified type where the user belongs to in the application
	 */
  public List<Group> getByMember(String appKey, User member, Integer type) {
		GroupSearchParamsBuilder builder = SearchParamsBuilder.group().member(member);
		if (type != null) builder.type(type);
		return search(appKey, builder.create());
	}

	/**
	 * Find all groups of the specified type where a user belongs to.
	 *
	 * @param memberId the identifier of the user
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @param offset the start position of the first result, numbered from 0
	 * @param max the maximum number of results to retrieve
	 * @return All groups of the specified type where the user belongs to in the application
	 */
	public List<Group> getByMember(int memberId, Integer type, Integer offset, Integer max) {
		GroupSearchParamsBuilder builder = SearchParamsBuilder.group().memberId(memberId).offset(offset).max(max);
		if (type != null) builder.type(type);
		return search(builder.create());
  }

	/**
	 * Find all groups of the specified type where a user belongs to within an
	 * application.
	 *
   * @param appKey the key of the application to search in
	 * @param memberId the identifier of the user
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @param offset the start position of the first result, numbered from 0
	 * @param max the maximum number of results to retrieve
	 * @return All groups of the specified type where the user belongs to in the application
	 */
	public List<Group> getByMember(String appKey, int memberId, Integer type, Integer offset, Integer max) {
		GroupSearchParamsBuilder builder = SearchParamsBuilder.group().memberId(memberId).offset(offset).max(max);
		if (type != null) builder.type(type);
		return search(appKey, builder.create());
  }

	/**
	 * Find a subset of groups of the specified type where a managed User entity
	 * belongs to.
	 *
	 * @param member the managed User entity
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @param offset the start position of the first result, numbered from 0
	 * @param max the maximum number of results to retrieve
	 * @return A subset of groups of the specified type where the user belongs to
	 */
	public List<Group> getByMember(User member, Integer type, Integer offset, Integer max) {
		GroupSearchParamsBuilder builder = SearchParamsBuilder.group().member(member).offset(offset).max(max);
		if (type != null) builder.type(type);
		return search(builder.create());
  }

	/**
	 * Find a subset of groups of the specified type where a managed User entity
	 * belongs to within an application. When offset is <code>null</code>,
	 * an offset of 0 will be used. When max is <code>null</code>, all results
	 * will be returned.
	 *
   * @param appKey the key of the application to search in
	 * @param member the managed User entity
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @param offset the start position of the first result, numbered from 0
	 * @param max the maximum number of results to retrieve
	 * @return A subset of groups of the specified type where the user belongs to in the application 
	 */
	public List<Group> getByMember(String appKey, User member, Integer type, Integer offset, Integer max) {
		GroupSearchParamsBuilder builder = SearchParamsBuilder.group().member(member).offset(offset).max(max);
		if (type != null) builder.type(type);
		return search(appKey, builder.create());
  }

	/**
	 * Count the number of child groups of a parent group that are of the
	 * specified type.
	 *
	 * @param parentId the identifier of the parent group
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @param depth how deep to search in the parent subtree
	 * @return the number of child groups of the parent group that are of the specified type
	 * @see #countByParent(java.lang.String, com.lodgon.dali.core.entity.Group, java.lang.Integer, java.lang.Integer)
	 */
	public Long countByParent(int parentId, Integer type, Integer depth) {
		GroupSearchParamsBuilder builder = SearchParamsBuilder.group().parentId(parentId).depth(depth);
		if (type != null) builder.type(type);
		return countSearch(builder.create());
	}

	/**
	 * Count the number of child groups of a parent group that are of the
	 * specified type within an application.
	 *
   * @param appKey the key of the application to search in
	 * @param parentId the identifier of the parent group
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @param depth how deep to search in the parent subtree
	 * @return the number of child groups of the parent group that are of the specified type in the application
	 * @see #countByParent(java.lang.String, com.lodgon.dali.core.entity.Group, java.lang.Integer, java.lang.Integer)
	 */
	public Long countByParent(String appKey, int parentId, Integer type, Integer depth) {
		GroupSearchParamsBuilder builder = SearchParamsBuilder.group().parentId(parentId).depth(depth);
		if (type != null) builder.type(type);
		return countSearch(appKey, builder.create());
	}

	/**
	 * Count the number of child groups of a managed parent Group entity that are
	 * of the specified type.
	 *
	 * @param parent the managed parent Group entity
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @param depth how deep to search in the parent subtree
	 * @return the number of child groups of the parent group that are of the specified type
	 * @see #countByParent(java.lang.String, com.lodgon.dali.core.entity.Group, java.lang.Integer, java.lang.Integer)
	 */
	public Long countByParent(Group parent, Integer type, Integer depth) {
		GroupSearchParamsBuilder builder = SearchParamsBuilder.group().parent(parent).depth(depth);
		if (type != null) builder.type(type);
		return countSearch(builder.create());
	}

	/**
	 * Count the number of child groups of a managed parent Group entity that are
	 * of the specified type within an application. The <code>depth</code> parameter
	 * specifies how deep to search in the parent's subtree. A value 1 or <code>null</code>
	 * means that it will only return the count of all the children that are
	 * directly below the parent. A value of 2 will include their children and so
	 * on. A negative value will search the entire subtree.
	 *
   * @param appKey the key of the application to search in
	 * @param parent the managed parent Group entity
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @param depth how deep to search in the parent subtree
	 * @return the number of child groups of the parent group that are of the specified type in the application
	 */
	public Long countByParent(String appKey, Group parent, Integer type, Integer depth) {
		GroupSearchParamsBuilder builder = SearchParamsBuilder.group().parent(parent).depth(depth);
		if (type != null) builder.type(type);
		return countSearch(appKey, builder.create());
	}

	/**
	 * Find all the child groups of a parent group that are of the
	 * specified type.
	 *
	 * @param parentId the identifier of the parent group
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @param depth how deep to search in the parent subtree
	 * @return All children of the parent group that are of the specified type
	 * @see #getByParent(java.lang.String, com.lodgon.dali.core.entity.Group, java.lang.Integer, java.lang.Integer, java.lang.Integer, java.lang.Integer)
	 */
  public List<Group> getByParent(int parentId, Integer type, Integer depth) {
		GroupSearchParamsBuilder builder = SearchParamsBuilder.group().parentId(parentId).depth(depth);
		if (type != null) builder.type(type);
		return search(builder.create());
  }

	/**
	 * Find all the child groups of a parent group that are of the
	 * specified type within an application.
	 *
   * @param appKey the key of the application to search in
	 * @param parentId the identifier of the parent group
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @param depth how deep to search in the parent subtree
	 * @return All children of the parent group that are of the specified type in the application
	 * @see #getByParent(java.lang.String, com.lodgon.dali.core.entity.Group, java.lang.Integer, java.lang.Integer, java.lang.Integer, java.lang.Integer)
	 */
  public List<Group> getByParent(String appKey, int parentId, Integer type, Integer depth) {
		GroupSearchParamsBuilder builder = SearchParamsBuilder.group().parentId(parentId).depth(depth);
		if (type != null) builder.type(type);
		return search(appKey, builder.create());
  }

	/**
	 * Find all the child groups of a managed parent group entity that
	 * are of the specified type.
	 *
	 * @param parent the managed parent Group entity
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @param depth how deep to search in the parent subtree
	 * @return All children of the parent group that are of the specified type
	 * @see #getByParent(java.lang.String, com.lodgon.dali.core.entity.Group, java.lang.Integer, java.lang.Integer, java.lang.Integer, java.lang.Integer)
	 */
  public List<Group> getByParent(Group parent, Integer type, Integer depth) {
		GroupSearchParamsBuilder builder = SearchParamsBuilder.group().parent(parent).depth(depth);
		if (type != null) builder.type(type);
		return search(builder.create());
  }

	/**
	 * Find all the child groups of a managed parent group entity that
	 * are of the specified type within an application.
	 *
   * @param appKey the key of the application to search in
	 * @param parent the managed parent Group entity
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @param depth how deep to search in the parent subtree
	 * @return All children of the parent group that are of the specified type in the application
	 * @see #getByParent(java.lang.String, com.lodgon.dali.core.entity.Group, java.lang.Integer, java.lang.Integer, java.lang.Integer, java.lang.Integer)
	 */
  public List<Group> getByParent(String appKey, Group parent, Integer type, Integer depth) {
		GroupSearchParamsBuilder builder = SearchParamsBuilder.group().parent(parent).depth(depth);
		if (type != null) builder.type(type);
		return search(appKey, builder.create());
  }

	/**
	 * Find a subset of the child groups of a parent group that are of the
	 * specified type.
	 *
	 * @param parentId the identifier of the parent group
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @param depth how deep to search in the parent subtree
	 * @param offset the start position of the first result, numbered from 0
	 * @param max the maximum number of results to retrieve
	 * @return a subset of children of the parent group that are of the specified type
	 * @see #getByParent(java.lang.String, com.lodgon.dali.core.entity.Group, java.lang.Integer, java.lang.Integer, java.lang.Integer, java.lang.Integer)
	 */
  public List<Group> getByParent(int parentId, Integer type, Integer depth, Integer offset, Integer max) {
		GroupSearchParamsBuilder builder = SearchParamsBuilder.group().parentId(parentId).depth(depth).offset(offset).max(max);
		if (type != null) builder.type(type);
		return search(builder.create());
  }

	/**
	 * Find a subset of the child groups of a parent group that are of the
	 * specified type within an application.
	 *
   * @param appKey the key of the application to search in
	 * @param parentId the identifier of the parent group
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @param depth how deep to search in the parent subtree
	 * @param offset the start position of the first result, numbered from 0
	 * @param max the maximum number of results to retrieve
	 * @return a subset of children of the parent group that are of the specified type in the application
	 * @see #getByParent(java.lang.String, com.lodgon.dali.core.entity.Group, java.lang.Integer, java.lang.Integer, java.lang.Integer, java.lang.Integer)
	 */
  public List<Group> getByParent(String appKey, int parentId, Integer type, Integer depth, Integer offset, Integer max) {
		GroupSearchParamsBuilder builder = SearchParamsBuilder.group().parentId(parentId).depth(depth).offset(offset).max(max);
		if (type != null) builder.type(type);
		return search(appKey, builder.create());
  }

	/**
	 * Find a subset of the child groups of a managed parent group entity that
	 * are of the specified type.
	 *
	 * @param parent the managed parent Group entity
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @param depth how deep to search in the parent subtree
	 * @param offset the start position of the first result, numbered from 0
	 * @param max the maximum number of results to retrieve
	 * @return a subset of children of the parent group that are of the specified type
	 * @see #getByParent(java.lang.String, com.lodgon.dali.core.entity.Group, java.lang.Integer, java.lang.Integer, java.lang.Integer, java.lang.Integer)
	 */
  public List<Group> getByParent(Group parent, Integer type, Integer depth, Integer offset, Integer max) {
		GroupSearchParamsBuilder builder = SearchParamsBuilder.group().parent(parent).depth(depth).offset(offset).max(max);
		if (type != null) builder.type(type);
		return search(builder.create());
  }

	/**
	 * Find a subset of the child groups of a managed parent group entity that are
	 * of the specified type within an application. When offset is <code>null</code>,
	 * an offset of 0 will be used. When max is <code>null</code>, all results
	 * will be returned. The <code>depth</code> parameter specifies how deep to
	 * search in the parent's subtree. A value 1 or <code>null</code> means that
	 * it will only return the count of all the children that are directly below
	 * the parent. A value of 2 will include their children and so on. A negative
	 * value will search the entire subtree.
	 *
   * @param appKey the key of the application to search in
	 * @param parent the managed parent Group entity
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @param depth how deep to search in the parent subtree
	 * @param offset the start position of the first result, numbered from 0
	 * @param max the maximum number of results to retrieve
	 * @return a subset of children of the parent group that are of the specified type in the application
	 * @see #getByParent(java.lang.String, com.lodgon.dali.core.entity.Group, java.lang.Integer, java.lang.Integer, java.lang.Integer, java.lang.Integer)
	 */
  public List<Group> getByParent(String appKey, Group parent, Integer type, Integer depth, Integer offset, Integer max) {
		GroupSearchParamsBuilder builder = SearchParamsBuilder.group().parent(parent).depth(depth).offset(offset).max(max);
		if (type != null) builder.type(type);
		return search(appKey, builder.create());
  }

	/**
	 * Count the number of groups of an author that are of the specified type.
	 *
	 * @param authorId the identifier of the author
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @return the number of groups of the author that are of the specified type
	 */
	public Long countByAuthor(int authorId, Integer type) {
		GroupSearchParamsBuilder builder = SearchParamsBuilder.group().authorId(authorId);
		if (type != null) builder.type(type);
		return countSearch(builder.create());
	}

	/**
	 * Count the number of groups of an author that are of the specified type
	 * within an application.
	 *
   * @param appKey the key of the application to search in
	 * @param authorId the identifier of the author
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @return the number of groups of the author that are of the specified type in the application
	 */
	public Long countByAuthor(String appKey, int authorId, Integer type) {
		GroupSearchParamsBuilder builder = SearchParamsBuilder.group().authorId(authorId);
		if (type != null) builder.type(type);
		return countSearch(appKey, builder.create());
	}

	/**
	 * Count the number of groups of a managed author entity that are
	 * of the specified type.
	 *
	 * @param author the managed User entity
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @return the number of groups of the author that are of the specified type
	 */
	public Long countByAuthor(User author, Integer type) {
		GroupSearchParamsBuilder builder = SearchParamsBuilder.group().author(author);
		if (type != null) builder.type(type);
		return countSearch(builder.create());
	}

	/**
	 * Count the number of groups of a managed author entity that are
	 * of the specified type within an application.
	 *
   * @param appKey the key of the application to search in
	 * @param author the managed User entity
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @return the number of groups of the author that are of the specified type in the application
	 */
	public Long countByAuthor(String appKey, User author, Integer type) {
		GroupSearchParamsBuilder builder = SearchParamsBuilder.group().author(author);
		if (type != null) builder.type(type);
		return countSearch(appKey, builder.create());
	}

	/**
	 * Find all groups of an author that are of the specified type.
	 *
   * @param authorId the identifier of the author
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @return All groups of the author that are of the specified type
	 */
  public List<Group> getByAuthor(int authorId, Integer type) {
		GroupSearchParamsBuilder builder = SearchParamsBuilder.group().authorId(authorId);
		if (type != null) builder.type(type);
		return search(builder.create());
  }

	/**
	 * Find all groups of an author that are of the specified type within an
	 * application.
	 *
   * @param appKey the key of the application to search in
	 * @param authorId the identifier of the author
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @return All groups of the author that are of the specified type in the application
	 */
  public List<Group> getByAuthor(String appKey, int authorId, Integer type) {
		GroupSearchParamsBuilder builder = SearchParamsBuilder.group().authorId(authorId);
		if (type != null) builder.type(type);
		return search(appKey, builder.create());
  }

	/**
	 * Find all groups of a managed author entity that are of the specified type.
	 *
	 * @param author the managed User entity
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @return All groups of the author that are of the specified type
	 */
  public List<Group> getByAuthor(User author, Integer type) {
		GroupSearchParamsBuilder builder = SearchParamsBuilder.group().author(author);
		if (type != null) builder.type(type);
		return search(builder.create());
  }

	/**
	 * Find all groups of a managed author entity that are of the specified type
	 * within an application.
	 *
   * @param appKey the key of the application to search in
	 * @param author the managed User entity
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @return All groups of the author that are of the specified type in the application
	 */
  public List<Group> getByAuthor(String appKey, User author, Integer type) {
		GroupSearchParamsBuilder builder = SearchParamsBuilder.group().author(author);
		if (type != null) builder.type(type);
		return search(appKey, builder.create());
  }

	/**
	 * Find a subset of groups of an author that are of the specified type.
	 *
   * @param authorId the identifier of the author
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @param offset the start position of the first result, numbered from 0
	 * @param max the maximum number of results to retrieve
	 * @return a subset of groups of the author that are of the specified type
	 */
  public List<Group> getByAuthor(int authorId, Integer type, Integer offset, Integer max) {
		GroupSearchParamsBuilder builder = SearchParamsBuilder.group().authorId(authorId).offset(offset).max(max);
		if (type != null) builder.type(type);
		return search(builder.create());
  }

	/**
	 * Find a subset of groups of an author that are of the specified type within an
	 * application.
	 *
   * @param appKey the key of the application to search in
	 * @param authorId the identifier of the author
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @param offset the start position of the first result, numbered from 0
	 * @param max the maximum number of results to retrieve
	 * @return a subset of groups of the author that are of the specified type in the application
	 */
  public List<Group> getByAuthor(String appKey, int authorId, Integer type, Integer offset, Integer max) {
		GroupSearchParamsBuilder builder = SearchParamsBuilder.group().authorId(authorId).offset(offset).max(max);
		if (type != null) builder.type(type);
		return search(appKey, builder.create());
  }

	/**
	 * Find a subset of groups of a managed author entity that are of the specified type.
	 *
	 * @param author the managed User entity
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @param offset the start position of the first result, numbered from 0
	 * @param max the maximum number of results to retrieve
	 * @return a subset of groups of the author that are of the specified type
	 */
  public List<Group> getByAuthor(User author, Integer type, Integer offset, Integer max) {
		GroupSearchParamsBuilder builder = SearchParamsBuilder.group().author(author).offset(offset).max(max);
		if (type != null) builder.type(type);
		return search(builder.create());
  }

	/**
	 * Find a subset of groups of a managed author entity that are of the
	 * specified type within an application. When offset is <code>null</code>,
	 * an offset of 0 will be used. When max is <code>null</code>, all results
	 * will be returned.
	 *
   * @param appKey the key of the application to search in
	 * @param author the managed User entity
	 * @param type the type to search for or <code>null</code> to search for all types
	 * @param offset the start position of the first result, numbered from 0
	 * @param max the maximum number of results to retrieve
	 * @return a subset of groups of the author that are of the specified type in the application
	 */
  public List<Group> getByAuthor(String appKey, User author, Integer type, Integer offset, Integer max) {
		GroupSearchParamsBuilder builder = SearchParamsBuilder.group().author(author).offset(offset).max(max);
		if (type != null) builder.type(type);
		return search(appKey, builder.create());
  }

	/**
	 * Add a user as a member to a group.
	 *
	 * @param groupId the identifier of the group
	 * @param userId the identifier of the user
	 * @return true when the user was added to the group, false otherwise
	 */
  public boolean join(int groupId, int userId) {
    Group group = em.find(Group.class, groupId);
    User user = em.find(User.class, userId);
    if (group != null && user != null) {
			return join(group, user);
    } else {
      return false;
    }
  }

	/**
	 * Add a managed User entity as a member to a managed Group entity.
	 *
	 * @param group the managed Group entity
	 * @param user the managed User entity
	 * @return true when the user was added to the group, false otherwise
	 */
  public boolean join(Group group, User user) {
		if (!group.getMembers().contains(user)) {
			group.getMembers().add(user);
			user.getGroups().add(group);
			return true;
		} else {
			return false;
		}
  }

	/**
	 * Remove a user as a member from a group.
	 *
	 * @param groupId the identifier of the group
	 * @param userId the identifier of the user
	 * @return true when the user was removed from the group, false otherwise
	 */
  public boolean leave(int groupId, int userId) {
    Group group = em.find(Group.class, groupId);
    User user = em.find(User.class, userId);
    if (group != null && user != null) {
			return leave(group, user);
    } else {
      return false;
    }
  }

	/**
	 * Remove a managed User entity as a member from a managed Group entity.
	 *
	 * @param group the managed Group entity
	 * @param user the managed User entity
	 * @return true when the user was removed from the group, false otherwise
	 */
  public boolean leave(Group group, User user) {
		if (group.getMembers().contains(user)) {
			group.getMembers().remove(user);
			user.getGroups().remove(group);
			return true;
		} else {
			return false;
		}
  }

  /**
   * Count the number of users that are a member of a group.
	 *
   * @param groupId the identifier of the group
   * @return the number of members of the group
   */
  public Long countMembers(int groupId) {
		return countMembers(em.find(Group.class, groupId));
  }

  /**
   * Count the number of users that are a member of a managed Group entity.
	 *
   * @param group the managed Group entity
   * @return the number of members of the group
   */
  public Long countMembers(Group group) {
    Query q = em.createNamedQuery("User.countInGroup");
    q.setParameter("group", group);
    return (Long) q.getSingleResult();
  }

	/**
	 * Find all members of a group.
	 *
	 * @param groupId the identifier of the group
	 * @return All members of the group
	 * @see #getMembers(com.lodgon.dali.core.entity.Group, java.lang.Integer, java.lang.Integer)
	 */
  public List<User> getMembers(int groupId) {
		return getMembers(groupId, null, null);
  }

	/**
	 * Find a subset of the members of a group. When offset is less than 0, the
	 * offset 0 will be used. When max is less than 0, all results will be
	 * returned.
	 *
	 * @param groupId the identifier of the group
	 * @param offset the start position of the first result, numbered from 0
	 * @param max the maximum number of results to retrieve
	 * @return A subset of the members of the group
	 * @see #getMembers(com.lodgon.dali.core.entity.Group, java.lang.Integer, java.lang.Integer)
	 */
  public List<User> getMembers(int groupId, Integer offset, Integer max) {
		return getMembers(em.find(Group.class, groupId), offset, max);
  }

	/**
	 * Find all members of a managed Group entity.
	 *
	 * @param group the managed Group entity
	 * @return All members of the group
	 * @see #getMembers(com.lodgon.dali.core.entity.Group, java.lang.Integer, java.lang.Integer)
	 */
  public List<User> getMembers(Group group) {
		return getMembers(group, null, null);
  }

	/**
	 * Find a subset of the members of a managed Group entity. When offset is less
	 * than 0, the offset 0 will be used. When max is less than 0, all results
	 * will be returned.
	 *
	 * @param group the managed Group entity
	 * @param offset the start position of the first result, numbered from 0
	 * @param max the maximum number of results to retrieve
	 * @return A subset of the members of the group
	 */
  public List<User> getMembers(Group group, Integer offset, Integer max) {
		Query q = em.createNamedQuery("User.findInGroup");
		q.setParameter("group", group);
		if (offset != null) q.setFirstResult(offset);
		if (max != null) q.setMaxResults(max);
		return q.getResultList();
  }

	/**
	 * Find out if a user belongs to a group.
	 *
	 * @param groupId the identifier of the group
	 * @param userId the identifier of the user
	 * @return true when the user belongs to the group, false otherwise
	 */
  public boolean isMember(int groupId, int userId) {
		return isMember(em.find(Group.class, groupId), em.find(User.class, userId));
  }

	/**
	 * Find out if a managed User entity belongs to a managed Group entity.
	 *
	 * @param group the managed Group entity
	 * @param user the managed User entity
	 * @return true when the user belongs to the group, false otherwise
	 */
  public boolean isMember(Group group, User user) {
		Query q = em.createNamedQuery("Group.isMember");
		q.setParameter("group", group);
		q.setParameter("user", user);
		return ! q.getResultList().isEmpty();
  }

  /**
   * Count the total number of groups.
	 *
   * @return the total number of groups
   */
	public Long countAll() {
		return countAll("");
	}

  /**
   * Count the total number of groups within an application.
	 *
   * @param appKey the key of the application to search in
   * @return the total number of groups in the application
   */
	public Long countAll(String appKey) {
		Query q = em.createNamedQuery("Group.count");
		q.setParameter("appKey", appKey);
		return (Long) q.getSingleResult();
	}

  /**
   * Count the number of groups that match the specified search parameters.
	 *
	 * @param params the search parameters that are used to generate the search query
   * @return the number of groups that match the search
   */
	public Long countSearch(GroupSearchParams params) {
		return countSearch("", params);
	}

  /**
   * Count the number of groups that match the specified search parameters
	 * within an application.
	 *
   * @param appKey the key of the application to search in
	 * @param params the search parameters that are used to generate the search query
   * @return the number of groups that match the search in the application
   */
	public Long countSearch(String appKey, GroupSearchParams params) {
		return searchProvider.count(appKey, params);
	}

	/**
	 * Finds all groups that match the specified search parameters.
	 *
	 * @param params the search parameters that are used to generate the search query
	 * @return all groups that match the specified search parameters
	 */
	public List<Group> search(GroupSearchParams params) {
		return search("", params);
	}

	/**
	 * Finds all groups that match the specified search parameters within an
	 * application.
	 *
   * @param appKey the key of the application to search in
	 * @param params the search parameters that are used to generate the search query
	 * @return all groups that match the specified search parameters in the application
	 */
	public List<Group> search(String appKey, GroupSearchParams params) {
		return searchProvider.search(appKey, params);
	}
}
