/*
 * Copyright (c) 2009-2011 LodgON  (http://www.lodgon.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package com.lodgon.dali.core.ejb;

import com.lodgon.dali.core.Constants;
import com.lodgon.dali.core.entity.Content;
import com.lodgon.dali.core.entity.Field;
import com.lodgon.dali.core.entity.Group;
import com.lodgon.dali.core.entity.Status;
import com.lodgon.dali.core.entity.TypePermission;
import com.lodgon.dali.core.entity.User;
import com.lodgon.dali.core.entity.UserPermission;
import com.lodgon.dali.core.entity.search.SearchParamsBuilder;
import com.lodgon.dali.core.entity.search.UserSearchParams;
import com.lodgon.dali.core.search.ActionType;
import com.lodgon.dali.core.search.DefaultUserSearchProvider;
import com.lodgon.dali.core.search.EntityAction;
import com.lodgon.dali.core.search.SearchProvider;
import java.util.List;
import java.util.UUID;
import java.util.logging.Level;
import javax.annotation.PostConstruct;
import javax.ejb.EJB;
import javax.ejb.Stateless;
import javax.enterprise.event.Event;
import javax.enterprise.inject.Any;
import javax.enterprise.inject.Instance;
import javax.inject.Inject;
import javax.persistence.EntityManager;
import javax.persistence.NoResultException;
import javax.persistence.NonUniqueResultException;
import javax.persistence.PersistenceContext;
import javax.persistence.Query;

/**
 * This class provides the persistence operations for users.
 *
 * @author erwin
 */
@Stateless
public class UserBean {
	@Inject @EntityAction(ActionType.CREATE)
	private Event<User> createUserEvent;
	@Inject @EntityAction(ActionType.UPDATE)
	private Event<User> updateUserEvent;
	@Inject @EntityAction(ActionType.DELETE)
	private Event<User> deleteUserEvent;

	@EJB
	AuthorizationBean authorizationBean;
	@EJB
	ContentBean contentBean;
	@EJB
	GroupBean groupBean;

	@Inject @Any
	private Instance<SearchProvider<User>> searchProviders;
	private SearchProvider<User> searchProvider;

	@PersistenceContext
	EntityManager em;

	@PostConstruct
	private void initializeContentBean() {
		if (searchProviders != null) {
			if (! searchProviders.isUnsatisfied() && ! searchProviders.isAmbiguous()) {
				searchProvider = searchProviders.get();
			} else if (searchProviders.isAmbiguous()) {
				for (SearchProvider<User> sp : searchProviders) {
					if (! (sp instanceof DefaultUserSearchProvider)) {
						searchProvider = sp;
					}
				}
			}
		}

		if (searchProvider == null) {
			Constants.LOGGER.log(Level.SEVERE, "No implementation for SearchProvider<User> was found.");
		} else {
			Constants.LOGGER.log(Level.INFO, "Using {0} as SearchProvider<User>.", searchProvider.getClass().getName());
		}
	}

	/**
	 * Add a User to persistent storage.
	 * If <code>user.getUid()</code> is not <code>null</code>, the value of it should be unique.
	 * If <code>user.getUid()</code> is null, this method will automatically generate a UUID.<br/>
	 * If <code>user.getStatus()</code> is null, the user will get the status
	 * {@link com.lodgon.dali.core.entity.Status#ACTIVE ACTIVE} by default.
	 *
	 * @param user the user that has to be made persistent.
	 * @return the persisted User object. An id is created for this User, hence
	 * <code>user.getId()</code> is not <code>null</code>
	 * @throws DaliCoreException when a conflict occurs with the appKey or when
	 * the provided <code>user.getUid()</code> is not unique
	 */
	public User create(User user) throws DaliCoreException {
		if (user.getAppKey() == null) {
			user.setAppKey("");
		}

		if (user.getFields() != null) {
			for (Field field : user.getFields()) {
				field.setUser(user);
			}
		}

		if (user.getUid() != null) {
			User existing = getByUid(user.getAppKey(), user.getUid());
			if (existing != null) {
				throw new DaliCoreException(DaliCoreException.UUID_CONFLICT, "user creation failed because " + user.getUid() + " already exists");
			}
		} else {
			user.setUid(UUID.randomUUID().toString());
		}

		if (user.getStatus() == null) {
			user.setStatus(Status.ACTIVE);
		}

		em.persist(user);
		createUserEvent.fire(user);
		return user;
	}

	/**
	 * Changes the password of a user.
	 *
	 * @param userId the identifier of the user
	 * @param password the new plain unhashed password
	 */
	public void setPassword(int userId, String password) {
		setPassword(em.find(User.class, userId), password);
	}

	/**
	 * Changes the password of a managed User entity.
	 *
	 * @param user the managed User entity
	 * @param password the new plain unhashed password
	 */
	public void setPassword(User user, String password) {
		if (user != null) {
			user.setPassword(HashUtil.getHash(password));
		}
	}

	/**
	 * Updates a user in the persistent storage. This method will <b>not</b>
	 * change the password or the groups of the user.
	 *
	 * @param user the user to update
   * @return the updated User object
	 */
	public User update(User user) {
		if (user.getAppKey() == null) {
			user.setAppKey("");
		}
		User dbuser = em.find(User.class, user.getId());
		user.setPassword(dbuser.getPassword());
		user.setGroups(dbuser.getGroups());

		if (user.getFields() != null) {
			for (Field f : user.getFields()) {
				f.setUser(user);
			}
		}

		user = em.merge(user);
		updateUserEvent.fire(user);
		return user;
	}

	/**
	 * Remove a user from persistent storage. This method removes the User in an
	 * unrecoverable way. For disabling a user, the <code>setStatus()</code>
	 * method should be used. The content owned by the user is not removed, but
	 * the author field of the content is set to <code>null</code>.
	 *
	 * @param userId the identifier of the user to remove
	 */
	public void delete(int userId) {
		delete(em.find(User.class, userId));
	}

	/**
	 * Remove a managed User entity from persistent storage. This method removes
	 * the User in an unrecoverable way. For disabling a user,
	 * {@link com.lodgon.dali.core.entity.User#setStatus(com.lodgon.dali.core.entity.Status)}
	 * should be used. The content owned by the user is not removed, but the
	 * author field of the content is set to <code>null</code>
	 *
	 * @param user the managed User entity to remove
	 */
	public void delete(User user) {
		if (user != null) {
			List<TypePermission> typePermissions = authorizationBean.getTypePermissionsByUser(user);
			for (TypePermission typePermission : typePermissions) {
				em.remove(typePermission);
			}
	    List<UserPermission> userPermissions = authorizationBean.getUserPermissionsByUser(user);
		  for (UserPermission userPermission : userPermissions) {
				em.remove(userPermission);
			}
	    List<UserPermission> targetUserPermissions = authorizationBean.getUserPermissionsByTarget(user);
		  for (UserPermission targetUserPermission : targetUserPermissions) {
				em.remove(targetUserPermission);
			}

			List<Content> contents = contentBean.getByAuthor(user.getAppKey(), user, null, null, null, 0, 0);
			for (Content content : contents) {
				content.setAuthor(null);
				em.merge(content);
			}

			List<Group> authorGroups = groupBean.getByAuthor(user.getAppKey(), user, null);
			for (Group group : authorGroups) {
				group.setAuthor(null);
				em.merge(group);
			}

			List<Group> memberGroups = groupBean.getByMember(user, null);
			for (Group group : memberGroups) {
				groupBean.leave(group, user);
			}

			deleteUserEvent.fire(user);
			em.remove(user);
		}
	}

	/**
	 * Find a user specified by the auto-created identifier
	 *
	 * @param userId the identifier of the user
   * @return the User that matches the specified identifier or null otherwise
	 */
	public User getById(int userId) {
		return em.find(User.class, userId);
	}

	/**
	 * Find a user specified by the unique identifier.
	 * Users with status {@link com.lodgon.dali.core.entity.Status#HIDDEN} will
	 * never be found using this method.
	 *
   * @param uid the unique identifier of the user
   * @return the User that matches the specified uid or null otherwise
	 * @throws DaliCoreException when more than one user was found with the specified uid
	 */
	public User getByUid(String uid) throws DaliCoreException {
		return getByUid("", uid);
	}

	/**
   * Find a user specified by the unique identifier within an application.
	 * Users with status {@link com.lodgon.dali.core.entity.Status#HIDDEN} will
	 * never be found using this method.
	 *
   * @param appKey the key of the application to search in
   * @param uid the unique identifier of the user
   * @return the User that matches the specified uid in the application or null otherwise
	 * @throws DaliCoreException when more than one user was found with the specified uid
	 */
	public User getByUid(String appKey, String uid) throws DaliCoreException {
		Query q = em.createNamedQuery("User.findByUid");
		q.setParameter("uid", uid);
		q.setParameter("appKey", appKey);
		q.setParameter("hidden", Status.HIDDEN);
		User user = null;
		try {
			user = (User) q.getSingleResult();
		} catch (NoResultException e) {
			Constants.LOGGER.log(Level.INFO, "No user found with uid: {0}", uid);
		} catch (NonUniqueResultException e) {
			Constants.LOGGER.log(Level.SEVERE, "More than one user found with uid: {0}", uid);
			throw new DaliCoreException(DaliCoreException.NON_UNIQUE, "non unique user result for " + uid);
		}
		return user;
	}

	/**
	 * Find a user with the specified email address.
	 * Users with status {@link com.lodgon.dali.core.entity.Status#HIDDEN} will
	 * never be found using this method.
	 *
	 * @param email the email address to look for
	 * @return the User that has this email address
	 * @throws DaliCoreException when more than one user was found with the specified email address
	 */
	public User getByEmail(String email) throws DaliCoreException {
		return getByEmail("", email);
	}

	/**
	 * Find a user with the specified email address within an application.
	 * Users with status {@link com.lodgon.dali.core.entity.Status#HIDDEN} will
	 * never be found using this method.
	 *
   * @param appKey the key of the application to search in
	 * @param email the email address to look for
	 * @return the User that has this email address
	 * @throws DaliCoreException when more than one user was found with the specified email address
	 */
	public User getByEmail(String appKey, String email) throws DaliCoreException {
		Query q = em.createNamedQuery("User.findByEmail");
		q.setParameter("email", email);
		q.setParameter("appKey", appKey);
		q.setParameter("hidden", Status.HIDDEN);
		User user = null;
		try {
			user = (User) q.getSingleResult();
		} catch (NoResultException e) {
			Constants.LOGGER.log(Level.INFO, "No user found with email: {0}", email);
		} catch (NonUniqueResultException e) {
			Constants.LOGGER.log(Level.SEVERE, "More than one user found with email: {0}", email);
			throw new DaliCoreException(DaliCoreException.NON_UNIQUE, "non unique user result for " + email);
		}
		return user;
	}

	/**
	 * Find a user with the specified screen name.
	 * Users with status {@link com.lodgon.dali.core.entity.Status#HIDDEN} will
	 * never be found using this method.
	 *
	 * @param screenName the screen name to look for
	 * @return the User that has this screen name
	 * @throws DaliCoreException when more than one user was found with the specified screen name
	 */
	public User getByScreenName(String screenName) throws DaliCoreException {
		return getByScreenName("", screenName);
	}

	/**
	 * Find a user with the specified screen name within an application.
	 * Users with status {@link com.lodgon.dali.core.entity.Status#HIDDEN} will
	 * never be found using this method.
	 *
   * @param appKey the key of the application to search in
	 * @param screenName the screen name to look for
	 * @return the User that has this screen name
	 * @throws DaliCoreException when more than one user was found with the specified screen name
	 */
	public User getByScreenName(String appKey, String screenName) throws DaliCoreException {
		Query q = em.createNamedQuery("User.findByScreenName");
		q.setParameter("screenName", screenName);
		q.setParameter("appKey", appKey);
		q.setParameter("hidden", Status.HIDDEN);
		User user = null;
		try {
			user = (User) q.getSingleResult();
		} catch (NoResultException e) {
			Constants.LOGGER.log(Level.INFO, "No user found with screenName: {0}", screenName);
		} catch (NonUniqueResultException e) {
			Constants.LOGGER.log(Level.SEVERE, "More than one user found with screenName: {0}", screenName);
			throw new DaliCoreException(DaliCoreException.NON_UNIQUE, "non unique user result for " + screenName);
		}
		return user;
	}

	/**
   * Validates the e-mail address and password. Validation is successful when a
	 * single user exists with the provided e-mail address and password that
	 * doesn't have its status set to
	 * {@link com.lodgon.dali.core.entity.Status#BLOCKED blocked}.
	 *
	 * @param email the e-mail address of the user to check
	 * @param password the plain unhashed password to check against the user's password
	 * @return a single unblocked user that matches the provided e-mail address and has the same
	 * password as the provided password
	 * @throws DaliCoreException when more than one user was found with the provided e-mail address
	 */
	public User validateEmailAndPassword(String email, String password) throws DaliCoreException {
		return validateEmailAndPassword("", email, password);
	}

	/**
   * Validates the e-mail address and password within an application. Validation
	 * is successful when a single user exists with the provided e-mail address
	 * and password that doesn't have its status set to
	 * {@link com.lodgon.dali.core.entity.Status#BLOCKED blocked}.
	 *
   * @param appKey the key of the application to validate in
	 * @param email the e-mail address of the user to check
	 * @param password the plain unhashed password to check against the user's password
	 * @return a single unblocked user within the application that matches the
	 * provided e-mail address and has the same password as the provided password
	 * @throws DaliCoreException when more than one user was found with the provided e-mail address
	 */
	public User validateEmailAndPassword(String appKey, String email, String password) throws DaliCoreException {
		Query q = em.createNamedQuery("User.validateByEmail");
		q.setParameter("appKey", appKey);
		q.setParameter("email", email);
		q.setParameter("blocked", Status.BLOCKED);
		try {
			User user = (User) q.getSingleResult();
			if (user != null && HashUtil.getHash(password).equals(user.getPassword())) {
				return user;
			}
		} catch (NoResultException e) {
			Constants.LOGGER.log(Level.INFO, "No user found with email: {0}", email);
		} catch (NonUniqueResultException e) {
			Constants.LOGGER.log(Level.WARNING, "More than one user found with email: {0}", email);
			throw new DaliCoreException(DaliCoreException.NON_UNIQUE, "non unique user result for " + email);
		}
		return null;
	}

	/**
   * Validates the screen name and password. Validation is successful when a
	 * single user exists with the provided screen name and password that
	 * doesn't have its status set to
	 * {@link com.lodgon.dali.core.entity.Status#BLOCKED blocked}.
	 *
	 * @param screenName the screen name of the user to check
	 * @param password the plain unhashed password to check against the user's password
	 * @return a single unblocked user that matches the provided screen name and has the same
	 * password as the provided password
	 * @throws DaliCoreException when more than one user was found with the provided screen name
	 */
	public User validateScreenNameAndPassword(String screenName, String password) throws DaliCoreException {
		return validateScreenNameAndPassword("", screenName, password);
	}

	/**
   * Validates the screen name and password within an application. Validation is
	 * successful when a single user exists with the provided screen name and
	 * password that doesn't have its status set to
	 * {@link com.lodgon.dali.core.entity.Status#BLOCKED blocked}.
	 *
   * @param appKey the key of the application to validate in
	 * @param screenName the screen name of the user to check
	 * @param password the plain unhashed password to check against the user's password
	 * @return a single unblocked user within the application that matches the
	 * provided screen name and has the same password as the provided password
	 * @throws DaliCoreException when more than one user was found with the provided screen name
	 */
	public User validateScreenNameAndPassword(String appKey, String screenName, String password) throws DaliCoreException {
		Query q = em.createNamedQuery("User.validateByScreenName");
		q.setParameter("appKey", appKey);
		q.setParameter("screenName", screenName);
		q.setParameter("blocked", Status.BLOCKED);
		try {
			User user = (User) q.getSingleResult();
			if (user != null && HashUtil.getHash(password).equals(user.getPassword())) {
				return user;
			}
		} catch (NoResultException e) {
			Constants.LOGGER.log(Level.INFO, "No user found with screen name: {0}", screenName);
		} catch (NonUniqueResultException e) {
			Constants.LOGGER.log(Level.WARNING, "More than one user found with screen name: {0}", screenName);
			throw new DaliCoreException(DaliCoreException.NON_UNIQUE, "non unique user result for " + screenName);
		}
		return null;
	}

	/**
   * Validates the uid and password. Validation is successful when a
	 * single user exists with the provided uid and password that
	 * doesn't have its status set to
	 * {@link com.lodgon.dali.core.entity.Status#BLOCKED blocked}.
	 *
	 * @param uid the uid of the user to check
	 * @param password the plain unhashed password to check against the user's password
	 * @return a single unblocked user that matches the provided uid and has the same
	 * password as the provided password
	 * @throws DaliCoreException when more than one user was found with the provided uid
	 */
	public User validateUidAndPassword(String uid, String password) throws DaliCoreException {
		return validateUidAndPassword("", uid, password);
	}

	/**
   * Validates the uid and password within an application. Validation is
	 * successful when a single user exists with the provided uid and password
	 * that doesn't have its status set to
	 * {@link com.lodgon.dali.core.entity.Status#BLOCKED blocked}.
	 *
   * @param appKey the key of the application to validate in
	 * @param uid the uid of the user to check
	 * @param password the plain unhashed password to check against the user's password
	 * @return a single unblocked user within the application that matches the
	 * provided uid and has the same password as the provided password
	 * @throws DaliCoreException when more than one user was found with the provided uid
	 */
	public User validateUidAndPassword(String appKey, String uid, String password) throws DaliCoreException {
		Query q = em.createNamedQuery("User.validateByUid");
		q.setParameter("appKey", appKey);
		q.setParameter("uid", uid);
		q.setParameter("blocked", Status.BLOCKED);
		try {
			User user = (User) q.getSingleResult();
			if (user != null && HashUtil.getHash(password).equals(user.getPassword())) {
				return user;
			}
		} catch (NoResultException e) {
			Constants.LOGGER.log(Level.INFO, "No user found with uid: {0}", uid);
		} catch (NonUniqueResultException e) {
			Constants.LOGGER.log(Level.WARNING, "More than one user found with uid: {0}", uid);
			throw new DaliCoreException(DaliCoreException.NON_UNIQUE, "non unique user result for " + uid);
		}
		return null;
	}

  /**
   * Count the number of users that match a specific type.
	 *
   * @param type the type of the user
   * @return the number of users that match the specified type
   */
  public Long countByType(int type) {
		return countSearch(SearchParamsBuilder.user().type(type).create());
  }

  /**
   * Count the number of users that match a specific type within an
	 * application.
	 *
   * @param appKey the key of the application to search in
   * @param type the type of the user
   * @return the number of users that match the specified type in the application
   */
  public Long countByType(String appKey, int type) {
		return countSearch(appKey, SearchParamsBuilder.user().type(type).create());
  }

	/**
	 * Find all users that are of the specified type.
	 *
	 * @param type the type to search for
	 * @return All users of the specified type
	 * @see #getByType(java.lang.String, int, java.lang.Integer, java.lang.Integer)
	 */
  public List<User> getByType(int type) {
		return search(SearchParamsBuilder.user().type(type).create());
  }

	/**
	 * Find all users that are of the specified type within an application.
	 *
   * @param appKey the key of the application to search in
	 * @param type the type to search for
	 * @return All users of the specified type in the application
	 * @see #getByType(java.lang.String, int, java.lang.Integer, java.lang.Integer)
	 */
  public List<User> getByType(String appKey, int type) {
		return search(appKey, SearchParamsBuilder.user().type(type).create());
  }

	/**
	 * Find a subset of users that are of the specified type.
	 *
	 * @param type the type to search for
	 * @param offset the start position of the first result, numbered from 0
	 * @param max the maximum number of results to retrieve
	 * @return A subset of users of the specified type
	 * @see #getByType(java.lang.String, int, java.lang.Integer, java.lang.Integer)
	 */
  public List<User> getByType(int type, Integer offset, Integer max) {
		return search(SearchParamsBuilder.user().type(type).offset(offset).max(max).create());
  }

	/**
	 * Find a subset of users that are of the specified type within an
	 * application. When offset is <code>null</code>, an offset of 0 will be used.
	 * When max is <code>null</code>, all results will be returned.
	 *
   * @param appKey the key of the application to search in
	 * @param type the type to search for
	 * @param offset the start position of the first result, numbered from 0
	 * @param max the maximum number of results to retrieve
	 * @return A subset of users of the specified type in the application
	 */
  public List<User> getByType(String appKey, int type, Integer offset, Integer max) {
		return search(appKey, SearchParamsBuilder.user().type(type).offset(offset).max(max).create());
  }

  /**
   * Count the number of users.
	 *
   * @return the total number of users
   */
	public Long countAll() {
		return countAll("");
	}

  /**
   * Count the number of users within an application. Users with status
	 * {@link com.lodgon.dali.core.entity.Status#HIDDEN} are not included in the
	 * count.
	 *
   * @param appKey the key of the application to search in
   * @return the total number of users in the application
   */
	public Long countAll(String appKey) {
		Query q = em.createNamedQuery("User.count");
		q.setParameter("appKey", appKey);
		q.setParameter("hidden", Status.HIDDEN);
		return (Long) q.getSingleResult();
	}

  /**
   * Count the number of users that match the specified search parameters.
	 *
	 * @param params the search parameters that are used to generate the search query
   * @return the total number of users that match the search
   */
	public Long countSearch(UserSearchParams params) {
		return countSearch("", params);
	}

  /**
   * Count the number of users that match the specified search parameters
	 * within an application.
	 *
   * @param appKey the key of the application to search in
	 * @param params the search parameters that are used to generate the search query
   * @return the total number of users that match the search in the application
   */
	public Long countSearch(String appKey, UserSearchParams params) {
		return searchProvider.count(appKey, params);
	}

	/**
	 * Finds all users that match the specified search parameters.
	 *
	 * @param params the search parameters that are used to generate the search query
	 * @return all users that match the specified search parameters
	 */
	public List<User> search(UserSearchParams params) {
		return search("", params);
	}

	/**
	 * Finds all users that match the specified search parameters within an
	 * application.
	 *
   * @param appKey the key of the application to search in
	 * @param params the search parameters that are used to generate the search query
	 * @return all users that match the specified search parameters in the application
	 */
	public List<User> search(String appKey, UserSearchParams params) {
		return searchProvider.search(appKey, params);
	}
}
