/*
 * Copyright (c) 2009-2011 LodgON  (http://www.lodgon.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package com.lodgon.dali.core.entity;

import java.io.Serializable;
import javax.persistence.Entity;
import javax.persistence.Id;
import javax.persistence.NamedQueries;
import javax.persistence.NamedQuery;
import javax.persistence.PrePersist;
import javax.xml.bind.annotation.XmlAccessType;
import javax.xml.bind.annotation.XmlAccessorType;
import javax.xml.bind.annotation.XmlRootElement;

/**
 * A DaliCore project can have a number of applications. Most of the methods
 * on the stateless session beans have an optional applicationKey parameter.
 * Doing so, similar projects can leverage the same DaliCore instance.
 * For simple projects, an Application is not needed. In that case, the
 * default Application, which has an empty String as its ApplicationKey
 * will be used.
 *
 * @author joeri
 */
@Entity
@NamedQueries({
  @NamedQuery(
    name = "Application.findAll",
    query = "SELECT a FROM Application a ORDER BY a.creationDate"
  )
})
@XmlRootElement
@XmlAccessorType(XmlAccessType.NONE)
public class Application implements Serializable {
  private static final long serialVersionUID = 1L;
  @Id
  private String appKey;
  private long creationDate;
  private String secretKey;

	/**
	 * The unique identifier of the application.
	 *
	 * @return unique identifier of the application
	 */
  public String getAppKey() {
    return appKey;
  }

	/**
	 * Defines the unique identifier of the application. This should be unique
	 * over all applications as it is also used as the unique identifier in the
	 * underlying persistence provider.
	 *
	 * @param appKey the unique identifier of the application
	 */
  public void setAppKey(String appKey) {
    this.appKey = appKey;
  }

	/**
	 * The time in milliseconds when the application was created.
	 *
	 * @return the time in milliseconds since midnight, January 1, 1970 UTC
	 */
  public long getCreationDate() {
    return creationDate;
  }

	/**
	 * Sets the time when the application was created. Note that the creation date
	 * will be automatically set if it was not provided when persisting the
	 * application.
	 *
	 * @param creationDate the time in milliseconds when the application was created
	 */
  public void setCreationDate(long creationDate) {
    this.creationDate = creationDate;
  }

	/**
	 * The secret key of the application.
	 *
	 * @return secret key of the application
	 */
  public String getSecretKey() {
    return secretKey;
  }

	/**
	 * Sets the secret key for the application. This is required when you want to
	 * {@link com.lodgon.dali.core.ejb.ApplicationBean#validateApplicationKey(java.lang.String, java.lang.String) validate an application}.
	 *
	 * @param secretKey the application's secret key
	 */
  public void setSecretKey(String secretKey) {
    this.secretKey = secretKey;
  }

  @PrePersist
  public void prePersist() {
    if (creationDate == 0) {
      this.creationDate = System.currentTimeMillis();
    }
  }
}
