/*
 * Copyright (c) 2009-2011 LodgON  (http://www.lodgon.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package com.lodgon.dali.core.entity;

import java.io.Serializable;
import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.GeneratedValue;
import javax.persistence.GenerationType;
import javax.persistence.Id;
import javax.persistence.ManyToOne;
import javax.persistence.NamedQueries;
import javax.persistence.NamedQuery;
import javax.persistence.PostLoad;
import javax.persistence.PostPersist;
import javax.persistence.PostUpdate;
import javax.persistence.Transient;
import javax.xml.bind.annotation.XmlAccessType;
import javax.xml.bind.annotation.XmlAccessorType;
import javax.xml.bind.annotation.XmlRootElement;

/**
 * A content permission is a permission that is granted to a user or group to
 * allow a certain action on a specific content entity. An example can be to
 * allow a user or group to remove (action) a specific content (target content).
 *
 * @author erwin
 */
@Entity
@NamedQueries({
  @NamedQuery(
    name = "ContentPermission.findByGroupAndContentAndAction",
    query = "SELECT cp FROM ContentPermission cp WHERE cp.group = :group AND cp.content = :content AND cp.action = :action"
  ),
  @NamedQuery(
    name = "ContentPermission.findByUserAndContentAndAction",
    query = "SELECT cp FROM ContentPermission cp WHERE cp.user = :user AND cp.content = :content AND cp.action = :action"
  ),
  @NamedQuery(
    name = "ContentPermission.findByGroupAndContent",
    query = "SELECT cp FROM ContentPermission cp WHERE cp.group = :group AND cp.content = :content"
  ),
  @NamedQuery(
    name = "ContentPermission.findByUserAndContent",
    query = "SELECT cp FROM ContentPermission cp WHERE cp.user = :user AND cp.content = :content"
  ),
  @NamedQuery(
    name = "ContentPermission.findByContent",
    query = "SELECT cp FROM ContentPermission cp WHERE cp.content = :content ORDER BY cp.id"
  ),
  @NamedQuery(
    name = "ContentPermission.findByGroup",
    query = "SELECT cp FROM ContentPermission cp WHERE cp.group = :group ORDER BY cp.id"
  ),
  @NamedQuery(
    name = "ContentPermission.findByUser",
    query = "SELECT cp FROM ContentPermission cp WHERE cp.user = :user ORDER BY cp.id"
  )
})
@XmlRootElement
@XmlAccessorType(XmlAccessType.NONE)
public class ContentPermission implements Serializable {
  private static final long serialVersionUID = 1L;
  @Id
  @GeneratedValue(strategy = GenerationType.IDENTITY)
  private Integer id;
  @ManyToOne
  private User user;
  @ManyToOne
  private Group group;
  @ManyToOne
  private Content content;
  @Column(name = "PERMISSIONACTION")
  private int action;

  @Transient
  private Integer contentId;

  @Transient
  private Authorizable authorizable;

	/**
	 * The action associated with this content permission.
	 *
	 * @return the action of the content permission
	 */
  public int getAction() {
    return action;
  }

	/**
	 * Sets the action associated with this content permission.
	 *
	 * @param action the action of the content permission
	 */
  public void setAction(int action) {
    this.action = action;
  }

	/**
	 * The group that was granted this content permission. If the permission was 
	 * instead granted to a user, then the group will be <code>null</code>.
	 *
	 * @return the group that was granted the content permission
	 */
  public Group getGroup() {
    return group;
  }

	/**
	 * Grants the content permission to the specified group.
	 *
	 * @param group the group that this content permission will be granted to
	 */
  public void setGroup(Group group) {
    this.group = group;
  }

	/**
	 * The unique identifier automatically generated by the persistence provider.
	 *
	 * @return unique identifier of the content permission
	 */
  public Integer getId() {
    return id;
  }

	/**
	 * Sets the unique identifier for the content permission. Note that this
	 * method should never be called by developers as it will be automatically set
	 * by the underlying persistence provider.
	 *
	 * @param id the unique identifier
	 */
  public void setId(Integer id) {
    this.id = id;
  }

	/**
	 * The user that was granted this content permission. If the permission was 
	 * instead granted to a group, then the user will be <code>null</code>.
	 *
	 * @return the user that was granted the content permission
	 */
  public User getUser() {
    return user;
  }

	/**
	 * Grants the content permission to the specified user.
	 *
	 * @param user the user that this content permission will be granted to
	 */
  public void setUser(User user) {
    this.user = user;
  }

	/**
	 * The content that this content permission was created for.
	 *
	 * @return the content for which the content permission is created
	 */
  public Content getContent() {
    return content;
  }

	/**
	 * Sets the content where this content permission is created for.
	 *
	 * @param content the content on which the content permission is created
	 */
  public void setContent(Content content) {
    this.content = content;
  }

	/**
	 * The id of the target content.
	 *
	 * @return the id the target content
	 */
  public Integer getContentId() {
    return contentId;
  }

	/**
	 * Sets the id of the target content. This method should not be called
	 * directly by developers as the field will be set automatically by taking the
	 * id of the target content field.
	 *
	 * @param contentId the id of the target content
	 */
  public void setContentId(Integer contentId) {
    this.contentId = contentId;
  }

	/**
	 * An authorizable that represents either the user or group that this
	 * content permission was granted to.
	 *
	 * @return an authorizable representing the entity this content permission was granted to
	 */
  public Authorizable getAuthorizable() {
    return authorizable;
  }

  @PostLoad
  @PostPersist
  @PostUpdate
  public void postLoad() {
    authorizable = new Authorizable();
    if (group != null) {
      authorizable.setId(group.getId());
			authorizable.setUid(group.getUid());
      authorizable.setType("group");
    }
    if (user != null) {
      authorizable.setId(user.getId());
			authorizable.setUid(user.getUid());
      authorizable.setType("user");
    }
    if (content != null) {
      setContentId(content.getId());
    }
  }
}
