/*
 * Copyright (c) 2009-2011 LodgON  (http://www.lodgon.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package com.lodgon.dali.core.entity.search;

import com.lodgon.dali.core.entity.Field;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

/**
 * Base class that defines search parameters that are generic for searching
 * the DaliCore Content, Group or User store.
 *
 * @author joeri
 */
public class SearchParams {
  private Long dateFrom;
  private Long dateTo;
	private List<Field> fields = new LinkedList<Field>();
  private Integer max;
  private Integer offset;
	private Order order = Order.ASCENDING;
	private Quantification quantification = Quantification.ALL;
  private String sort;
  private Set<Integer> types = new HashSet<Integer>();

	/**
	 * The date in milliseconds to search from.
	 *
	 * @return the date in milliseconds to search from
	 */
	public Long getDateFrom() {
		return dateFrom;
	}

	/**
	 * Defines the date in milliseconds to search from. All entities that were
	 * created after this date will be returned.
	 *
	 * @param dateFrom 
	 */
	public void setDateFrom(Long dateFrom) {
		this.dateFrom = dateFrom;
	}

	/**
	 * The date in milliseconds to search to.
	 *
	 * @return the date in milliseconds to search to
	 */
	public Long getDateTo() {
		return dateTo;
	}

	/**
	 * Defines the date in milliseconds to search to. All entities that were
	 * created before this date will be returned.
	 *
	 * @param dateTo
	 */
	public void setDateTo(Long dateTo) {
		this.dateTo = dateTo;
	}

	/**
	 * The list of fields to search by.
	 *
	 * @return the list of fields to search by
	 */
	public List<Field> getFields() {
		return fields;
	}

	/**
	 * Defines the list of fields to search by. Only fields that match both the
	 * key and value of a field will be returned in the search results. You can
	 * use * as a wildcard, but only in the field values.
	 *
	 * @param fields a list of fields to search by
	 */
	public void setFields(List<Field> fields) {
		this.fields = fields;
	}

	/**
	 * The maximum number of search results that will be returned.
	 *
	 * @return the maximum number of results returned
	 */
	public Integer getMax() {
		return max;
	}

	/**
	 * Defines the maximum number of search results that should be returned. A
	 * value of <code>null</code>, which is also the default value, will return
	 * all possible results.
	 *
	 * @param max the maximum number of results to return
	 */
	public void setMax(Integer max) {
		this.max = max;
	}

	/**
	 * The start position of the first result, numbered from 0.
	 *
	 * @return the start position of the first result
	 */
	public Integer getOffset() {
		return offset;
	}

	/**
	 * Defines the start position of the first result, numbered from 0. A value
	 * of <code>null</code>, which is also the default value, is the same as
	 * specifying an offset of 0.
	 *
	 * @param offset the start position of the first result
	 */
	public void setOffset(Integer offset) {
		this.offset = offset;
	}

	/**
	 * The order in which the results will be sorted.
	 *
	 * @return the order of the sort
	 */
	public Order getOrder() {
		return order;
	}

	/**
	 * Defines the order in which the results should be sorted. By default the
	 * results will be returned in ascending order. A value of <code>null</code>
	 * will also return the results in ascending order.
	 *
	 * @param order the sort order
	 */
	public void setOrder(Order order) {
		this.order = order;
	}

	/**
	 * Returns Quantification.ALL if an entity needs to match all the specified
	 * search parameters or Quantification.ANY if one parameter is enough.
	 *
	 * @return the quantification to apply to the search parameters
	 */
	public Quantification getQuantification() {
		return quantification;
	}

	/**
	 * Defines whether the to be searched entities should match either all or any
	 * of the specified search parameters. By default all the search parameters
	 * need to be matched in order to be returned in the results.
	 *
	 * @param quantification the quantification to apply to the search parameters
	 */
	public void setQuantification(Quantification quantification) {
		this.quantification = quantification;
	}

	/**
	 * The name of the field to sort the results by.
	 *
	 * @return the name of the field to sort by
	 */
	public String getSort() {
		return sort;
	}

	/**
	 * Defines by which field the search results should be sorted. The allowed
	 * sort values are specific to each entity. By default, the sort will occur
	 * by the creationDate of the entities.
	 *
	 * @param sort the name of the field to sort by
	 */
	public void setSort(String sort) {
		this.sort = sort;
	}

	/**
	 * The list of types to search by.
	 *
	 * @return the list of types to search by
	 */
	public Set<Integer> getTypes() {
		return types;
	}

	/**
	 * Defines the list of types to search by.
	 *
	 * @param types the list of types to search by
	 */
	public void setTypes(Set<Integer> types) {
		this.types = types;
	}

	/**
	 * Defines the possible values for the order of sorting search results.
	 */
	public enum Order {
		/**
		 * The search results are returned in ascending order. This is the default
		 * sort order.
		 */
		ASCENDING,
		/**
		 * The search results are returned in descending order.
		 */
		DESCENDING
	}

	/**
	 * Defines the possible values for the quantification that can be applied
	 * on the specified search parameters.
	 */
	public enum Quantification {
		/**
		 * When using ANY as the quantifier, all entities that match at least one of
		 * the specified search parameters will be returned in the search results.
		 */
		ANY,
		/**
		 * When using ALL as the quantifier, only the entities that match all of the
		 * specified search parameters will be returned in the search results. This
		 * is the default quantifier.
		 */
		ALL
	}
}
