/*
 * Copyright (c) 2009-2011 LodgON  (http://www.lodgon.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package com.lodgon.dali.core.search;

import com.lodgon.dali.core.entity.Content;
import com.lodgon.dali.core.entity.Field;
import com.lodgon.dali.core.entity.Group;
import com.lodgon.dali.core.entity.User;
import com.lodgon.dali.core.entity.search.ContentSearchParams;
import com.lodgon.dali.core.entity.search.SearchParams;
import java.util.LinkedList;
import java.util.List;
import javax.persistence.EntityManager;
import javax.persistence.PersistenceContext;
import javax.persistence.TypedQuery;
import javax.persistence.criteria.CriteriaBuilder;
import javax.persistence.criteria.CriteriaQuery;
import javax.persistence.criteria.Expression;
import javax.persistence.criteria.Join;
import javax.persistence.criteria.JoinType;
import javax.persistence.criteria.Predicate;
import javax.persistence.criteria.Root;

/**
 * This is the default implementation of SearchProvider for Content entities.
 * When no other SearchProvider implementation was found through CDI injection,
 * then DaliCore will fall back to this implementation, which performs its
 * searches by building JPA queries with the Criteria API.
 *
 * @author Joeri
 */
public class DefaultContentSearchProvider implements SearchProvider<Content> {
	@PersistenceContext
	EntityManager em;

	@Override
	public Long count(String appKey, SearchParams searchParams) {
		ContentSearchParams params = (ContentSearchParams) searchParams;

		// find out if an ID was provided of a non existing entity. when there is
		// one found, return a count of zero
		if (params.getAuthorId() != null) {
			User author = em.find(User.class, params.getAuthorId());
			if (author != null) {
				params.setAuthor(author);
			} else {
				return Long.valueOf(0);
			}
		}
		if (params.getGroupId() != null) {
			Group group = em.find(Group.class, params.getGroupId());
			if (group != null) {
				params.setGroup(group);
			} else {
				return Long.valueOf(0);
			}
		}
		if (params.getParentId() != null) {
			Content parent = em.find(Content.class, params.getParentId());
			if (parent != null) {
				params.setParent(parent);
			} else {
				return Long.valueOf(0);
			}
		}

		CriteriaBuilder builder = em.getCriteriaBuilder();
		CriteriaQuery<Long> query = builder.createQuery(Long.class);

		Root<Content> _content = query.from(Content.class);

		Expression where = createWhereExpression(appKey, builder, _content, params);

		query.select(builder.countDistinct(_content.get("id"))).where(where);

		TypedQuery<Long> q = em.createQuery(query);
		return q.getSingleResult();
	}

	@Override
	public List<Content> search(String appKey, SearchParams searchParams) {
		ContentSearchParams params = (ContentSearchParams) searchParams;

		// find out if an ID was provided of a non existing entity. when there is
		// one found, return an empty list
		if (params.getAuthorId() != null) {
			User author = em.find(User.class, params.getAuthorId());
			if (author != null) {
				params.setAuthor(author);
			} else {
				return new LinkedList<Content>();
			}
		}
		if (params.getGroupId() != null) {
			Group group = em.find(Group.class, params.getGroupId());
			if (group != null) {
				params.setGroup(group);
			} else {
				return new LinkedList<Content>();
			}
		}
		if (params.getParentId() != null) {
			Content parent = em.find(Content.class, params.getParentId());
			if (parent != null) {
				params.setParent(parent);
			} else {
				return new LinkedList<Content>();
			}
		}

		CriteriaBuilder builder = em.getCriteriaBuilder();
		CriteriaQuery<Integer> query = builder.createQuery(Integer.class);

		Root<Content> _content = query.from(Content.class);

		Expression where = createWhereExpression(appKey, builder, _content, params);

		query.distinct(true).select(_content.<Integer>get("id")).where(where);

		Expression sort = _content.get("creationDate");
		if (params.getSort() != null &&
						(params.getSort().equals("weight") ||
						params.getSort().equals("title") ||
						params.getSort().equals("creationDate") ||
						params.getSort().equals("modificationDate"))) {
			sort = _content.get(params.getSort());
		}

		if (params.getOrder() == null || params.getOrder() == SearchParams.Order.ASCENDING) {
			query.orderBy(builder.asc(sort), builder.asc(_content.get("id")));
		} else {
			query.orderBy(builder.desc(sort), builder.desc(_content.get("id")));
		}

		TypedQuery<Integer> q = em.createQuery(query);
		if (params.getOffset() != null) q.setFirstResult(params.getOffset());
		if (params.getMax() != null) q.setMaxResults(params.getMax());

		List<Content> contents = new LinkedList<Content>();
		List<Integer> contentIds = q.getResultList();
		for (Integer contentId : contentIds) {
			contents.add(em.find(Content.class, contentId));
		}
		return contents;
	}

	private Expression createWhereExpression(String appKey, CriteriaBuilder builder, Root<Content> _content, ContentSearchParams params) {
		SearchParams.Quantification quantor = params.getQuantification();

		Expression where = null;
		if (params.getContents() != null) {
			Expression whereContents = null;
			for (String content : params.getContents()) {
	      whereContents = addExpression(builder, whereContents, builder.like(_content.<String>get("content"), content.replaceAll("\\*", "%")), SearchParams.Quantification.ANY);
			}
			if (whereContents != null) {
				where = addExpression(builder, where, whereContents, quantor);
			}
    }
		if (params.getDateFrom() != null) {
			where = addExpression(builder, where, builder.gt(_content.<Long>get("creationDate"), params.getDateFrom()), quantor);
		}
		if (params.getDateTo() != null) {
			where = addExpression(builder, where, builder.lt(_content.<Long>get("creationDate"), params.getDateTo()), quantor);
		}
		if (params.getSummaries() != null) {
			Expression whereSummaries = null;
			for (String summary : params.getSummaries()) {
				whereSummaries = addExpression(builder, whereSummaries, builder.like(_content.<String>get("summary"), summary.replaceAll("\\*", "%")), SearchParams.Quantification.ANY);
			}
			if (whereSummaries != null) {
				where = addExpression(builder, where, whereSummaries, quantor);
			}
		}
		if (params.getTitles() != null) {
			Expression whereTitles = null;
			for (String title : params.getTitles()) {
				whereTitles = addExpression(builder, whereTitles, builder.like(_content.<String>get("title"), title.replaceAll("\\*", "%")), SearchParams.Quantification.ANY);
			}
			if (whereTitles != null) {
				where = addExpression(builder, where, whereTitles, quantor);
			}
    }

		if (params.getAuthor() != null) {
			where = addExpression(builder, where, builder.equal(_content.<User>get("author"), params.getAuthor()), quantor);
		}

		if (params.getGroup() != null) {
			where = addExpression(builder, where, builder.equal(_content.<Group>get("group"), params.getGroup()), quantor);
		}

		if (params.getParent() != null) {
			if (params.getDepth() == null || params.getDepth() == 1) {
				where = addExpression(builder, where, builder.equal(_content.<User>get("parent"), params.getParent()), quantor);
			} else {
				Predicate pathPredicate = builder.like(_content.<String>get("path"), (params.getParent().getPath().isEmpty() ? "/" + params.getParent().getId() : params.getParent().getPath() + params.getParent().getId()) + "/%");
				if (params.getDepth() < 0) {
					where = addExpression(builder, where, pathPredicate, quantor);
				} else {
					where = addExpression(builder, where, builder.and(pathPredicate, builder.le(_content.<Integer>get("depth"), params.getParent().getDepth() + params.getDepth())), quantor);
				}
			}
		}

		if (params.getTags() != null && ! params.getTags().isEmpty()) {
			for (String tag : params.getTags()) {
				Join<Content, String> _tags = _content.joinSet("tags", JoinType.LEFT);
				where = addExpression(builder, where, builder.like(_tags, tag.replaceAll("\\*", "%")), quantor);
			}
		}

    if (params.getFields() != null && ! params.getFields().isEmpty()) {
      for (Field field : params.getFields()) {
	      Join<Content, Field> _fields = _content.joinList("fields", JoinType.LEFT);
        Predicate fieldsPredicate = builder.and(builder.equal(_fields.get("name"), field.getName()), builder.like(_fields.<String>get("value"), field.getValue().replaceAll("\\*", "%")));
        where = addExpression(builder, where, fieldsPredicate, quantor);
      }
    }

		if (params.getTypes() != null && ! params.getTypes().isEmpty()) {
			where = addExpression(builder, where, _content.get("type").in(params.getTypes()), SearchParams.Quantification.ALL);
		}

		return addExpression(builder, where, builder.equal(_content.get("appKey"), appKey != null ? appKey : ""), SearchParams.Quantification.ALL);
	}

	private Expression addExpression(CriteriaBuilder builder, Expression where, Expression expression, SearchParams.Quantification quantification) {
		if (where == null) {
			return expression;
		} else {
			if (quantification == SearchParams.Quantification.ANY) {
				return builder.or(where, expression);
			} else {
				return builder.and(where, expression);
			}
		}
	}
}
