/*
 * Copyright (c) 2009-2011 LodgON  (http://www.lodgon.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package com.lodgon.dali.core.search;

import com.lodgon.dali.core.entity.Field;
import com.lodgon.dali.core.entity.Group;
import com.lodgon.dali.core.entity.Status;
import com.lodgon.dali.core.entity.User;
import com.lodgon.dali.core.entity.search.SearchParams;
import com.lodgon.dali.core.entity.search.UserSearchParams;
import java.util.LinkedList;
import java.util.List;
import javax.persistence.EntityManager;
import javax.persistence.PersistenceContext;
import javax.persistence.TypedQuery;
import javax.persistence.criteria.CriteriaBuilder;
import javax.persistence.criteria.CriteriaQuery;
import javax.persistence.criteria.Expression;
import javax.persistence.criteria.Join;
import javax.persistence.criteria.JoinType;
import javax.persistence.criteria.Predicate;
import javax.persistence.criteria.Root;

/**
 * This is the default implementation of SearchProvider for User entities.
 * When no other SearchProvider implementation was found through CDI injection,
 * then DaliCore will fall back to this implementation, which performs its
 * searches by building JPA queries with the Criteria API.
 *
 * @author Joeri
 */
public class DefaultUserSearchProvider implements SearchProvider<User> {
	@PersistenceContext
  EntityManager em;

	@Override
	public Long count(String appKey, SearchParams searchParams) {
		UserSearchParams params = (UserSearchParams) searchParams;

		// find out if an ID was provided of a non existing entity.
		if (params.getGroupIds() != null && ! params.getGroupIds().isEmpty()) {
			for (Integer groupId : params.getGroupIds()) {
				Group group = em.find(Group.class, groupId);
				if (group != null) {
					params.getGroups().add(group);
				}
			}
		}

		CriteriaBuilder builder = em.getCriteriaBuilder();
		CriteriaQuery<Long> query = builder.createQuery(Long.class);

		Root<User> _user = query.from(User.class);

		Expression where = createWhereExpression(appKey, builder, _user, params);

		query.select(builder.countDistinct(_user)).where(where);

		TypedQuery<Long> q = em.createQuery(query);
		return q.getSingleResult();
	}

	@Override
	public List<User> search(String appKey, SearchParams searchParams) {
		UserSearchParams params = (UserSearchParams) searchParams;

		// find out if an ID was provided of a non existing entity. when there is
		// one found, return an empty list
		if (params.getGroupIds() != null && ! params.getGroupIds().isEmpty()) {
			for (Integer groupId : params.getGroupIds()) {
				Group group = em.find(Group.class, groupId);
				if (group != null) {
					params.getGroups().add(group);
				}
			}
		}

		CriteriaBuilder builder = em.getCriteriaBuilder();
		CriteriaQuery<User> query = builder.createQuery(User.class);

		Root<User> _user = query.from(User.class);

		Expression where = createWhereExpression(appKey, builder, _user, params);

		query.distinct(true).select(_user).where(where);

		Expression sort = _user.get("creationDate");
		if (params.getSort() != null
						&& (params.getSort().equals("screenName")
						|| params.getSort().equals("firstName")
						|| params.getSort().equals("lastName")
						|| params.getSort().equals("type")
						|| params.getSort().equals("creationDate")
						|| params.getSort().equals("lastLoginDate"))) {
			sort = _user.get(params.getSort());
		}

		if (params.getOrder() == null || params.getOrder() == SearchParams.Order.ASCENDING) {
			query.orderBy(builder.asc(sort), builder.asc(_user.get("id")));
		} else {
			query.orderBy(builder.desc(sort), builder.asc(_user.get("id")));
		}

		TypedQuery<User> q = em.createQuery(query);
		if (params.getOffset() != null) {
			q.setFirstResult(params.getOffset());
		}
		if (params.getMax() != null) {
			q.setMaxResults(params.getMax());
		}
		return q.getResultList();
	}

	private Expression createWhereExpression(String appKey, CriteriaBuilder builder, Root<User> _user, UserSearchParams params) {
		SearchParams.Quantification quantor = params.getQuantification();

		Expression where = null;
		if (params.getDateFrom() != null) {
			where = addExpression(builder, where, builder.gt(_user.<Long>get("creationDate"), params.getDateFrom()), quantor);
		}
		if (params.getDateTo() != null) {
			where = addExpression(builder, where, builder.lt(_user.<Long>get("creationDate"), params.getDateTo()), quantor);
		}
		if (params.getEmails() != null) {
			Expression whereEmails = null;
			for (String email : params.getEmails()) {
				whereEmails = addExpression(builder, whereEmails, builder.like(_user.<String>get("email"), email.replaceAll("\\*", "%")), SearchParams.Quantification.ANY);
			}
			if (whereEmails != null) {
				where = addExpression(builder, where, whereEmails, quantor);
			}
		}
		if (params.getFirstNames() != null) {
			Expression whereFirstNames = null;
			for (String firstName : params.getFirstNames()) {
				whereFirstNames = addExpression(builder, whereFirstNames, builder.like(_user.<String>get("firstName"), firstName.replaceAll("\\*", "%")), SearchParams.Quantification.ANY);
			}
			if (whereFirstNames != null) {
				where = addExpression(builder, where, whereFirstNames, quantor);
			}
		}
		if (params.getLastNames() != null) {
			Expression whereLastNames = null;
			for (String lastName : params.getLastNames()) {
				whereLastNames = addExpression(builder, whereLastNames, builder.like(_user.<String>get("lastName"), lastName.replaceAll("\\*", "%")), SearchParams.Quantification.ANY);
			}
			if (whereLastNames != null) {
				where = addExpression(builder, where, whereLastNames, quantor);
			}
		}
		if (params.getScreenNames() != null) {
			Expression whereScreenNames = null;
			for (String screenName : params.getScreenNames()) {
				whereScreenNames = addExpression(builder, whereScreenNames, builder.like(_user.<String>get("screenName"), screenName.replaceAll("\\*", "%")), SearchParams.Quantification.ANY);
			}
			if (whereScreenNames != null) {
				where = addExpression(builder, where, whereScreenNames, quantor);
			}
		}

		if (params.getStatuses() != null && !params.getStatuses().isEmpty()) {
			List<String> statuses = new LinkedList<String>();
			for (Status status : params.getStatuses()) {
				statuses.add(status.getStatus());
			}
			where = addExpression(builder, where, _user.get("status").get("status").in(statuses), quantor);
		}

		if (params.getFields() != null && !params.getFields().isEmpty()) {
			for (Field field : params.getFields()) {
				Join<User, Field> _fields = _user.joinList("fields", JoinType.LEFT);
				Predicate fieldsPredicate = builder.and(builder.equal(_fields.get("name"), field.getName()), builder.like(_fields.<String>get("value"), field.getValue().replaceAll("\\*", "%")));
				where = addExpression(builder, where, fieldsPredicate, quantor);
			}
		}

		if (params.getGroups() != null && !params.getGroups().isEmpty()) {
			for (Group group : params.getGroups()) {
				Join<User, Group> _groups = _user.joinList("groups", JoinType.LEFT);
				where = addExpression(builder, where, builder.equal(_groups, group), SearchParams.Quantification.ALL);
			}
		}

		if (params.getTypes() != null && ! params.getTypes().isEmpty()) {
			where = addExpression(builder, where, _user.get("type").in(params.getTypes()), SearchParams.Quantification.ALL);
		}

		// if statuses were not queried, exclude HIDDEN users by default
		if (params.getStatuses() == null || params.getStatuses().isEmpty()) {
			where = addExpression(builder, where, builder.notEqual(_user.get("status"), Status.HIDDEN), SearchParams.Quantification.ALL);
		}

		return addExpression(builder, where, builder.equal(_user.get("appKey"), appKey != null ? appKey : ""), SearchParams.Quantification.ALL);
	}

	private Expression addExpression(CriteriaBuilder builder, Expression where, Expression expression, SearchParams.Quantification quantification) {
		if (where == null) {
			return expression;
		} else {
			if (quantification == SearchParams.Quantification.ANY) {
				return builder.or(where, expression);
			} else {
				return builder.and(where, expression);
			}
		}
	}
}
